import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;

import topology.Bridge;
import topology.FabanMetaInformation;
import topology.FabanWaveTool;
import topology.Link;
import topology.Topology;

public class TopologyIteratorWrapper {

	private int nodes;
	private int edges;

	// Example: nodes = 5
	//
	// __ N1 N2 N3 N4 N5
	// N1 -- e0 e1 e2 e3
	// N2 -- -- e4 e5 e6
	// N3 -- -- -- e7 e8
	// N4 -- -- -- -- e9
	// N5 -- -- -- -- --
	//
	// lookuptable_____ = (e9,e8,e7,e6,e5,e4,e3,e2,e1,e0)
	// edgelookupvector = ( 1, 0, 1, 0, 0, 1, 1, 0, 0, 1)
	//
	// N1-->N2-->N3-->N4-->N5-|
	// ^______________________|
	//
	// ==> Ring

	private boolean[] edgeLookupVector; // Consist of 0.5n*(n-1) elements
	private int[][] lookuptable; // Analogous to the lookup table in the example above, where each edge e is a
									// two dimensional integer value, e.g. e = e[0]-->e[1]

	private int lnNoOfIterations; // Specifies the limit for iterations, e.g. if the most k bits are set fixed,
									// then only all left bits are iterated, i.e. itLimit = edge - |msb|
	private int noOfOnes; // Holds the number of '1' in the edgeLookupVector without the fixed '1' in the msb
	private int msbOnes;
	private int progress = 0; // Holds the progress of the iterations and is a value between 0 and 127, where
								// the progress is

	private Topology topology;

	// Statistics
	private int noOfValidTopologies = 0;
	private double noOfValidTopologies_2pow31 = 0;
	private HashMap<Integer, HashMap<Integer, Long>> hopEdgeCorr;
	private ArrayList<Long> timeRecords;

	public TopologyIteratorWrapper(int nodes) {
		this.nodes = nodes;
		this.edges = nodes * (nodes - 1) / 2;
		this.noOfOnes = 0;
		this.topology = null;

		hopEdgeCorr = new HashMap<Integer, HashMap<Integer, Long>>();
		edgeLookupVector = new boolean[edges];
		lookuptable = new int[edges][2];
		timeRecords = new ArrayList<Long>();
		for (int k = edges - 1, i = 1; i <= nodes; i++) {
			for (int j = i + 1; j <= nodes; j++) {
				edgeLookupVector[k] = false;
				lookuptable[k][0] = i;
				lookuptable[k--][1] = j;
			}
		}
		lnNoOfIterations = edges;
	}

	public TopologyIteratorWrapper(int nodes, String msb) {
		this(nodes);
		msbOnes = 0;
		for (int i = 0; i < msb.length(); i++) {
			if (msb.charAt(i) == '1') {
				edgeLookupVector[i] = true;
				msbOnes++;
			}
		}
		lnNoOfIterations = edges - msb.length();
	}

	public int getNodes() {
		return nodes;
	}

	public int getEdges() {
		return edges;
	}

	public int getLnNumberOfIterations() {
		return lnNoOfIterations;
	}

	public void setLnNumberOfIterations(int no) {
		lnNoOfIterations = no;
	}

	public int getNoOfValidTopologies() {
		return noOfValidTopologies;
	}

	public double getNoOfValidTopologies_2pow31() {
		return noOfValidTopologies_2pow31;
	}

	public void next() {
		for (int i = edgeLookupVector.length - 1; i >= 0; i--) {
			if (edgeLookupVector[i]) {
				edgeLookupVector[i] = false;
				if (noOfOnes < lnNoOfIterations)
					noOfOnes--;
				else if (noOfOnes == lnNoOfIterations) {
					noOfOnes = lnNoOfIterations + 1;
					return;
				}
			} else {
				edgeLookupVector[i] = true;
				noOfOnes++;
				if (edgeLookupVector.length - 1 - i >= lnNoOfIterations - 7) {
					progress++;
					System.out.println("[" + Thread.currentThread().getName() + "] "
							+ (100 * progress) / (int) (lnNoOfIterations >= 7 ? 128 : Math.pow(2, lnNoOfIterations))
							+ "%");
				}
				return;
			}
		}
		topology = null;
	}

	public void nextRandom() {
		noOfOnes = 0;
		for (int i = edgeLookupVector.length - 1; i >= 0; i--) {
			if (Math.random() < TopoGenerator.edgeProb) {
				edgeLookupVector[i] = true;
				noOfOnes++;
			} else {
				edgeLookupVector[i] = false;
			}
		}
	}

	public void nextRandom(int ones) {
		noOfOnes = 0;
		for (int i = 0; i < edgeLookupVector.length; i++) {
			edgeLookupVector[i] = false;
		}
		Random rand = new Random();
		for (int i = 0; i < ones; i++) {
			while (true) {
				int ind = rand.nextInt(edges);
				if (!edgeLookupVector[ind]) {
					edgeLookupVector[ind] = true;
					noOfOnes++;
					break;
				}
			}

		}
	}

	public void generateTopology() {
		topology = new Topology();
		for (int i = 0; i < nodes; i++) {
			topology.addBridge("N" + (i + 1));
		}
		for (int i = 0; i < edges; i++) {
			if (edgeLookupVector[i]) {
				topology.addLink("N" + lookuptable[i][0], "N" + lookuptable[i][1]);
			}
		}
	}

	public boolean proveFaban(String selectionMode) {
		FabanWaveTool tool = new FabanWaveTool(topology, selectionMode);
		long timeRec = System.currentTimeMillis();
		boolean ret = tool.wavesExists();
		timeRecords.add(System.currentTimeMillis() - timeRec);
		if (ret) {
			if (noOfValidTopologies == Integer.MAX_VALUE) {
				noOfValidTopologies = 0;
				noOfValidTopologies_2pow31++;
			} else {
				noOfValidTopologies++;
			}

			int maxDist = 0;
			for (FabanMetaInformation fabmeta : tool.getNetwork().getWaves().values()) {
				maxDist = Math.max(maxDist,
						Math.max(fabmeta.getW1().getMaxDistance(), fabmeta.getW2().getMaxDistance()));
			}
			if (!hopEdgeCorr.containsKey(maxDist)) {
				hopEdgeCorr.put(maxDist, new HashMap<Integer, Long>());
			}
			HashMap<Integer, Long> m = hopEdgeCorr.get(maxDist);
			if (!m.containsKey(noOfOnes)) {
				m.put(noOfOnes + msbOnes, 1l);
			} else {
				m.put(noOfOnes + msbOnes, (Long) m.get(noOfOnes) + 1);
			}

		}

		return ret;
	}

	public double getValidTopoPercentage() {
		double val = Math.pow(2,
				((Math.log(noOfValidTopologies_2pow31 + ((double) noOfValidTopologies) / Integer.MAX_VALUE)
						/ Math.log(2) + 31) - lnNoOfIterations));
		return Math.round(1000000 * val) / 1000000.0;
	}

	public String boolArrToString() {
		String ret = "";
		for (boolean b : edgeLookupVector) {
			ret += (b ? "1" : "0");
		}
		return ret;
	}

	public void printTopology() {
		System.out.print(boolArrToString() + "\t");

		for (Bridge b : topology.getBridges()) {
			System.out.print(b.name + "  ");
		}
		System.out.print("--> {");
		for (Link l : topology.getBridgeLinks()) {
			System.out.print("(" + l.a.name + "," + l.b.name + ")");
		}
		System.out.println("}");
	}

	public boolean end() {
		return noOfOnes == lnNoOfIterations + 1;
	}

	public void writeStatisticsToFile(String filename) {
		try {
			FileWriter fw = new FileWriter(new File(filename + "_EdgeHopCorr.csv"));
			fw.write("maxhops\tedges\t#topologies\n");
			for (Integer maxDist : hopEdgeCorr.keySet()) {
				for (Integer ones : hopEdgeCorr.get(maxDist).keySet()) {
					fw.write(maxDist + "\t" + ones + "\t" + hopEdgeCorr.get(maxDist).get(ones) + "\n");
				}
			}
			fw.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		if(TopoGenerator.recordRuntimes) {
			try {
				FileWriter fw = new FileWriter(new File(filename + "_Runtimes.csv"));
				fw.write("Time [ms]\n");
				for (Long time: timeRecords) {
					fw.write(time + "\n");
				}
				fw.close();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

	}
	
	public void clearStatistics() {
		hopEdgeCorr.clear();
	}

}
