package configgui.java.commands.simple;

import java.io.File;
import java.io.FilenameFilter;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelSftp;
import com.jcraft.jsch.JSch;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.jcraft.jsch.SftpException;

import configgui.java.Rcc;
import configgui.java.commands.Command;
import configgui.java.model.configuration.BananaInformation;
import configgui.java.util.terminal.Terminal;
import javafx.collections.ObservableList;

public class UploadCommand extends Command {

	private ObservableList<BananaInformation> bananas;
	private String relativePath;
	int[] progress = { 0, 0 };

	public UploadCommand(ObservableList<BananaInformation> bananas, String opt) {
		this.bananas = bananas;
		this.relativePath = opt;
	}

	@Override
	protected void executeCommand() {
		if (relativePath == null) {
			File file = new File(Rcc.configuration.getSourcePath());
			if (file == null || !file.exists() || !file.isDirectory()) {
				Rcc.out.terminalPrintln("Source path not set properly", 1);
				success = false;
				return;
			}
			String[] directories = file.list(new FilenameFilter() {
				@Override
				public boolean accept(File current, String name) {
					return new File(current, name).isDirectory();
				}
			});
			final String[] dirs = directories;

			Rcc.out.terminalPrintln("Please specify a directory to upload", 1);
			int i = 0;
			for (String s : dirs) {
				if (++i < dirs.length)
					Rcc.out.terminalPrintln(Terminal.CMD_INDENT_CHARS + "\u2560\u2550 " + s, 1);
				else
					Rcc.out.terminalPrintln(Terminal.CMD_INDENT_CHARS + "\u255A\u2550 " + s, 1);

			}

			return;
		}

		if (Rcc.configuration.getDestinationPath() == null || Rcc.configuration.getDestinationPath().isEmpty()) {
			Rcc.out.terminalPrintln("Destination path not set", 1);
			return;
		}

		File uploadFolder = null;
		String[] relativePathSplitted = relativePath.split("/");
		String subpath = "";
		for (int i = 0; i < relativePathSplitted.length - 1; i++) {
			subpath += "/" + relativePathSplitted[i];
		}
		String foldername = relativePathSplitted[relativePathSplitted.length - 1];

		if (Rcc.configuration.getSourcePath() != null) {
			String srcPath = Rcc.configuration.getSourcePath() + subpath + "/" + foldername;
			srcPath.replaceAll("//", "/");
			srcPath.replaceAll("//", "/");
			uploadFolder = new File(srcPath);
		}
		if (uploadFolder == null || !uploadFolder.exists() || !uploadFolder.isDirectory()) {
			Rcc.out.terminalPrintln("No such file or folder", 1);
			success=false;
			return;
		}
		final File uF = uploadFolder;

		final String src = uploadFolder.getAbsolutePath();
		String dst = (Rcc.configuration.getDestinationPath() + subpath).replaceAll("//", "/");
		Rcc.out.terminalPrintln("Starting upload...\n\tFrom: \"" + src + "\"\n\t  To: \"" + dst + "\"\n ", 1);
		Rcc.out.activateProgressBar();
		Rcc.out.statusSetText("Upload started");
		progress[0] = progress[1] = 0;
		Rcc.out.updateProgressBar(progress[0], bananas.size());
		Rcc.out.lockTerminal();
		for (BananaInformation banana : bananas) {
			new Thread(new Runnable() {
				public void run() {
					String host = banana.getIP();
					Thread.currentThread().setName("SFTP Uploader: " + host);
					Rcc.out.setState(banana, BananaInformation.state_busy);
					Rcc.out.terminalPrintln("Upload to " + host + " started.", 1);
					JSch ssh = new JSch();
					Session session = null;
					ChannelSftp sftp = null;
					try {
						int totalProgress = bananas.size();
						Rcc.out.bananaOutPrintln(banana, "Upload started");
						session = ssh.getSession(Rcc.configuration.getRemoteUser(), host);
						session.setPassword(Rcc.configuration.getRemotePassword());
						session.setConfig("StrictHostKeyChecking", "no");
						session.connect();
						Channel channel = session.openChannel("sftp");
						channel.connect();
						sftp = (ChannelSftp) channel;
						boolean start = true;
						for (String folder : dst.split("/")) {
							if (start) {
								folder = "/" + folder;
								start = false;
							}
							if (folder.length() > 0) {
								try {
									sftp.cd(folder);
								} catch (SftpException e) {
									sftp.mkdir(folder);
									sftp.cd(folder);
								}
							}
						}
						File sourceFolder = uF;
						sendFiles(sftp, sourceFolder, banana);
						sftp.exit();
						session.disconnect();
						synchronized (bananas) {
							progress[0]++;
							Rcc.out.updateProgressBar(progress[0], totalProgress);
							Rcc.out.terminalPrintln("Upload to " + host + " finished.", 1);
							Rcc.out.bananaOutPrintln(banana, "Upload finished");
							bananas.notifyAll();
						}

					} catch (JSchException | SftpException e) {
						if (sftp != null && sftp.isConnected())
							sftp.exit();
						if (session != null && session.isConnected())
							session.disconnect();
						Rcc.out.addException(e);
						synchronized (bananas) {
							progress[1]++;
							Rcc.out.terminalPrintln("Upload to " + host + " failed.", 1);
							Rcc.out.bananaOutPrintln(banana, "Upload failed");
							bananas.notifyAll();
						}

					}
					Rcc.out.setState(banana, BananaInformation.state_idle);
				}
			}).start();
		}

		int totalProgress = bananas.size();
		while (progress[0] + progress[1] < totalProgress) {
			synchronized (bananas) {
				try {
					bananas.wait();
				} catch (Exception e) {
					Rcc.out.addException(e);
				}
			}
		}
		Rcc.out.statusSetText("Upload finished");
		Rcc.out.terminalPrintln("Upload finished.", 1);

		Rcc.out.unlockTerminal();
		return;
	}

	@Override
	public boolean validate() {
		return (bananas != null && Rcc.out != null && Rcc.configuration != null);
	}

	private void sendFiles(ChannelSftp sftp, File file, BananaInformation banana) throws SftpException {
		if (file != null && file.isDirectory()) {
			if (!remoteDirectoryExists(sftp, file)) {
				Rcc.out.bananaOutPrintln(banana, "Creating remote directory " + file.getName());
				sftp.mkdir(file.getName());
			}
			sftp.cd(file.getName());
			for (File f : file.listFiles())
				sendFiles(sftp, f, banana);
			sftp.cd("..");
		} else {
			if (remoteDirectoryExists(sftp, file))
				sftp.rm(file.getName());
			Rcc.out.bananaOutPrintln(banana, Terminal.CMD_INDENT_CHARS + "Uploading file " + file.getName());
			sftp.put(file.getAbsolutePath(), file.getName());
		}
	}

	private boolean remoteDirectoryExists(ChannelSftp sftp, File file) {
		try {
			sftp.lstat(file.getName());
		} catch (SftpException e) {
			if (e.id == ChannelSftp.SSH_FX_NO_SUCH_FILE) {
				return false;
			} else {
				Rcc.out.addException(e);
				return false;
			}
		}
		return true;
	}

}
