package configgui.java.util.terminal;

import java.util.ArrayList;

import org.fxmisc.richtext.PopupAlignment;

import configgui.java.controller.BananaController;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyCodeCombination;

/**
 * A subclass of ShCodeArea implementing a simple Terminal-like Node which sends
 * Commands to another class.
 */
public class Terminal extends ShCodeArea {

	/*
	 * Static Configuration Variables
	 */
	public static final String CMD_START = "> ";
	public static final String CMD_INDENT_CHARS = "  ";
	public static final int CMD_HISTORY_LENGTH = 1000;

	/*
	 * Connections to other Objects
	 */
	private BananaController controller;
	private AutoCompletionContextBox acPopup;

	/*
	 * Command handling
	 */
	private ArrayList<String> commandHistory = new ArrayList<String>();
	private int commandHistoryPointer = 0;

	/*
	 * Execution locking
	 */
	private boolean locked = false;

	/**
	 * Getter and setter functions
	 */
	public BananaController getController() {
		return controller;
	}

	public void setController(BananaController controller) {
		this.controller = controller;
	}

	public void setCompleteCmds(ArrayList<String> cmds) {
		acPopup.setCmds(cmds);
	}

	public Terminal() {

		super();

		acPopup = new AutoCompletionContextBox(Terminal.this);
		commandHistory.add("");

		/*
		 * Mouse handling
		 */
		setOnMouseReleased(e -> {
			if (getInput() != null) {
				int minPos = getText().length() - getInput().length();
				if (getCaretPosition() < minPos) {
					insertText(minPos, "");
				}
			}
		});

		/*
		 * Key handling
		 */
		setOnKeyReleased(e -> {
			if (getInput() != null) {
				int minPos = getText().length() - getInput().length();
				if (getCaretPosition() < minPos) {
					insertText(minPos, "");
				}
			}
		});
		setOnKeyTyped(e -> {
			if (locked || getInput() == null || !getInput().contains(getSelectedText()))
				e.consume();
		});

		setOnKeyPressed(e -> {

			if (e.getCode().equals(KeyCode.C) && e.isControlDown()) {
				this.appendText("\n> ");
				this.setEditable(true);
			} else if (getInput() != null) {
				/*
				 * Show Popupwindow after Keycombination CTRL+SPACE is pressed
				 */
				if (new KeyCodeCombination(KeyCode.SPACE, KeyCodeCombination.CONTROL_DOWN).match(e) || e.getCode().equals(KeyCode.TAB)) {
					int caretPosition = this.getCaretPosition();
					if (0 != caretPosition) {
						this.positionCaret(caretPosition - 1);
						this.positionCaret(caretPosition);
					} else {
						this.positionCaret(caretPosition + 1);
						this.positionCaret(caretPosition);
					}
					this.setPopupWindow(acPopup.getPopUp());
					this.setPopupAlignment(PopupAlignment.CARET_BOTTOM);
					acPopup.show(getInput());
					e.consume();
				}
				/*
				 * Read and dispatch current command line
				 */
				else if (e.getCode().equals(KeyCode.ENTER)) {
					String cmd = getInput();
					appendText("\n");
					controller.dispatchCommand(cmd);
					commandHistory.set(commandHistory.size() - 1, cmd);
					addCommandToHistory("");
					resetCommandHistoryPointer();
					e.consume();
				}
				/*
				 * Delete previous character in the current command line
				 */
				else if (e.getCode().equals(KeyCode.BACK_SPACE)) {
					if (getCaretColumn() <= CMD_START.length() || !getInput().contains(getSelectedText()))
						e.consume();
				}
				/*
				 * Delete next character in the current command line
				 */
				else if (e.getCode().equals(KeyCode.DELETE)) {
					if (!getInput().contains(getSelectedText()))
						e.consume();
				}
				/*
				 * Replace current command line by the previous command from command history
				 */
				else if (e.getCode().equals(KeyCode.UP)) {
					if (!commandHistory.isEmpty()) {
						if (commandHistoryPointer == commandHistory.size() - 1)
							commandHistory.set(commandHistory.size() - 1, getInput());
						deleteInput();
						commandHistoryPointer = Math.max(0, commandHistoryPointer - 1);
						this.appendText(getCommandFromHistory(commandHistoryPointer));
					}
					e.consume();
				}
				/*
				 * Replace current command line by the next command from command history
				 */
				else if (e.getCode().equals(KeyCode.DOWN)) {
					if (!commandHistory.isEmpty()) {
						if (commandHistoryPointer == commandHistory.size() - 1)
							commandHistory.set(commandHistory.size() - 1, getInput());
						deleteInput();
						commandHistoryPointer = Math.min(commandHistory.size() - 1, commandHistoryPointer + 1);
						this.appendText(getCommandFromHistory(commandHistoryPointer));
					}
					e.consume();
				} else {
					refreshPopup(getInput() + e.getText());
					resetCommandHistoryPointer();
				}
			}
		});
	}

	public String getInput() {
		String completeText = this.getText();
		int beginOfCommand = this.getText().lastIndexOf(CMD_START);
		if (completeText.lastIndexOf('\n') < beginOfCommand)
			return this.getText(beginOfCommand + CMD_START.length(), this.getText().length()).trim();
		else
			return null;
	}

	public void deleteInput() {
		this.deleteText(getText().length() - getInput().length(), getText().length());
	}
	/*
	 * Internal Methods
	 */

	private void addCommandToHistory(String s) {
		if (commandHistory.size() == CMD_HISTORY_LENGTH)
			commandHistory.remove(0);
		commandHistory.add(s);
		commandHistoryPointer = commandHistory.size() - 1;
	}

	private String getCommandFromHistory(int index) {
		return commandHistory.get(index);
	}

	private void refreshPopup(String s) {
		boolean stillValidInput = false;
		for (String cmd : acPopup.getCmds()) {
			if (cmd.equals(s))
				acPopup.hide();
			if (cmd.contains(s))
				stillValidInput = true;
		}
		if (!stillValidInput)
			acPopup.hide();
	}

	private void resetCommandHistoryPointer() {
		commandHistoryPointer = commandHistory.size() - 1;
	}

	/*
	 * External Methods
	 */

	public void sendCompletedCmd(String cmd) {
		deleteInput();
		this.appendText(cmd);
	}

	public boolean isLocked() {
		return locked;
	}

	public synchronized void lockTerminal() {
		this.locked = true;
	}

	public synchronized void unlockTerminal() {
		this.locked = false;
	}
}