//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#include "egressPort.h"
#include "../../packets/fabMsg_m.h"

namespace fabansimulator::bridge {

Define_Module(EgressPort);

void EgressPort::initialize() {
    _bridgeadmin = dynamic_cast<BridgeAdmin*>(this->getParentModule()->getSubmodule("bridgeAdmin"));
}

void EgressPort::handleMessage(cMessage *msg) {
    if (dynamic_cast<fabansimulator::FabMsg*>(msg) != nullptr) {
        FabMsg *fm = dynamic_cast<fabansimulator::FabMsg*>(msg);
        const int maxHop = (int) par("faultassumption") + 1;

        if (par("dupDetectionEgressPort").boolValue() && fm->getFaultinjections() == 0) {
            Data *data = dynamic_cast<Data*>(fm->decapsulate());
            uint32_t crc = _bridgeadmin->generateCrc32(fm, data);
            fm->encapsulate(data);
            if (_accHistoryContains(fm->getSenderID(), fm->getCheckingBridgeID(), crc) && par("faultassumption").intValue() > 1) {
                delete (msg);
                return;
            }
            else {
                _addAccHistory(fm->getSenderID(), fm->getCheckingBridgeID(), crc);
            }
        }

        if (_bridgeadmin->isDistributingBridge(fm)) {
            // Mode DB
            fm->setHopCounter((fm->getHopCounter() < maxHop ? fm->getHopCounter() + 1 : maxHop));
            fm->setSignature(fm->getSignature() ^ _bridgeadmin->getNetworkAdmin()->getDmask());
        }
        else if (_bridgeadmin->isPrecheckingBridge(fm)) {

            if (fm->getHopCounter() < par("faultassumption").intValue()) {
                fm->setSignature(fm->getSignature() ^ _bridgeadmin->getNetworkAdmin()->getPmask());
            }
            else {

            }

            fm->setHopCounter((fm->getHopCounter() < maxHop ? fm->getHopCounter() + 1 : maxHop));
        }
        else if (_bridgeadmin->isCheckingBridge(fm)) {
            fm->setHopCounter((fm->getHopCounter() < maxHop ? fm->getHopCounter() + 1 : maxHop));
            fm->setSignature(fm->getSignature() ^ _bridgeadmin->getNetworkAdmin()->getCmask());
            fm->setRoutingPhase(1);
        }
        else if (fm->getHopCounter() > (int) par("faultassumption") || fm->getRoutingPhase() == 1) {
            fm->setHopCounter((fm->getHopCounter() < maxHop ? fm->getHopCounter() + 1 : maxHop));
        }
        else {
            throw cRuntimeError("Invalid hop counter. Check implementation of \"brDec\" module!");
        }
        _totalThroughputBits += fm->getBitLength();
    }
    else {
        throw cRuntimeError("Unknown message received!");
    }

    send(msg, "out");
}

void EgressPort::finish() {
    recordScalar("Throughput", (_totalThroughputBits / 1000000.0) / simTime().dbl(), "Mbps");
}

void EgressPort::_addAccHistory(uint16_t senderId, uint16_t cbId, uint32_t sgn) {
    uint32_t id = (((uint32_t) senderId) << 16) + cbId;
    auto sid = accHistory.find(id);
    if (sid != accHistory.end()) {
        if (sid->second.size() < 5) {
            sid->second.push_back(sgn);
        }
        else {
            sid->second.erase(sid->second.begin());
            sid->second.push_back(sgn);
        }
    }
    else {
        std::vector<uint32_t> neuVec;
        neuVec.push_back(sgn);
        accHistory.insert(std::pair<uint32_t, std::vector<uint32_t>>(id, neuVec));
    }
}

bool EgressPort::_accHistoryContains(uint16_t senderId, uint16_t cbId, uint32_t sgn) {
    uint32_t id = (((uint32_t) senderId) << 16) + cbId;

    auto sid = accHistory.find(id);
    if (sid != accHistory.end()) {
        for (auto y = sid->second.begin(); y != sid->second.end(); y++) {
            if (sgn == *y) {
                return true;
            }
        }
    }
    return false;
}

BridgeAdmin* EgressPort::getBridgeadmin() {
    return _bridgeadmin;
}

}
//namespace
