//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#include "faultinjector.h"
#include "../../packets/fabMsg_m.h"
#include "../../packets/data_m.h"
#include "../../FabanSimLib.h"

namespace fabansimulator::bridge {

Define_Module(Faultinjector);

void Faultinjector::initialize() {
    _bridgeadmin = dynamic_cast<BridgeAdmin*>(this->getParentModule()->getSubmodule("bridgeAdmin"));
    if (par("active").boolValue()) {
        _bridgeadmin->setFailActive();
    }

    // Check Parameters
    if (par("failurePropCrash").doubleValue() < 0 || par("failurePropCrash").doubleValue() > 1) {
        throw cRuntimeError("Parameter \"failurePropCrash\" has an invalid value.");
    }
    if (par("failurePropOmission").doubleValue() < 0 || par("failurePropOmission").doubleValue() > 1) {
        throw cRuntimeError("Parameter \"failurePropOmission\" has an invalid value.");
    }
    if (par("failurePropDelay").doubleValue() < 0 || par("failurePropDelay").doubleValue() > 1) {
        throw cRuntimeError("Parameter \"failurePropDelay\" has an invalid value.");
    }
    if (par("failurePropBitflip").doubleValue() < 0 || par("failurePropBitflip").doubleValue() > 1) {
        throw cRuntimeError("Parameter \"failurePropBitflip\" has an invalid value.");
    }
    if (par("failurePropFieldManipulation").doubleValue() < 0 || par("failurePropFieldManipulation").doubleValue() > 1) {
        throw cRuntimeError("Parameter \"failurePropFieldManipulation\" has an invalid value.");
    }
    if (par("failurePropDuplication").doubleValue() < 0 || par("failurePropDuplication").doubleValue() > 1) {
        throw cRuntimeError("Parameter \"failurePropDuplication\" has an invalid value.");
    }
    if (par("failurePropWrongForwarding").doubleValue() < 0 || par("failurePropWrongForwarding").doubleValue() > 1) {
        throw cRuntimeError("Parameter \"failurePropWrongForwarding\" has an invalid value.");
    }
    if (par("failurePropBabblingComponent").doubleValue() < 0 || par("failurePropBabblingComponent").doubleValue() > 1) {
        throw cRuntimeError("Parameter \"failurePropBabblingComponent\" has an invalid value.");
    }
    if (par("failurePropAddValidSignManipulation").doubleValue() < 0 || par("failurePropAddValidSignManipulation").doubleValue() > 1) {
        throw cRuntimeError("Parameter \"failurePropAddValidSignManipulation\" has an invalid value.");
    }
}

void Faultinjector::refreshDisplay() const {
    char buf[40];
    sprintf(buf, "%dO/%dD", _numOmission, _numDelay);
    getDisplayString().setTagArg("t", 0, buf);
}

void Faultinjector::handleMessage(cMessage *msg) {
    if (par("active").boolValue()) {
        if (strcmp(msg->getName(), "BabblingComponentSelfMessage") == 0) {
            _bcCount--;
            delete (msg);

            FabMsg *fm = new FabMsg("", 0);
            fm->setSenderID((uint16_t) rand());
            fm->setSeqNo((uint8_t) rand());
            double d = (double) random();
            fm->setDeliveryTime((d >= 0 ? d : -1 * d));
            fm->setHopCounter((uint8_t) rand());
            fm->setSignature(rand());
            fm->setBitLength(fm->getPacketSizeInBits());
            Data *data = new Data("", 0);
            int datasize = (rand() / RAND_MAX) * 1373;
            data->setDataArraySize(datasize);
            data->setByteLength(datasize);
            for (unsigned int i = 0; i < data->getDataArraySize(); i++) {
                data->setData(i, (uint8_t) rand());
            }
            fm->encapsulate(data);
            fm->setFaultinjections(fm->getFaultinjections() + 1);
            send(fm, "out");
        }
        else if (dynamic_cast<FabMsg*>(msg) != nullptr) {
            // Store known SenderIds
            FabMsg *fm = dynamic_cast<fabansimulator::FabMsg*>(msg);
            bool n = true;
            for (auto sid : _knownSenderIds) {
                if (fm->getSenderID() == sid) {
                    n = false;
                }
            }
            if (n) {
                _knownSenderIds.push_back(fm->getSenderID());
            }

            // ##############
            // FAULTINJECTION
            // ##############

            // (1) Ausfallfehler
            if (strcmp("faultinjector", this->getName()) == 0 && (rand() < RAND_MAX * par("failurePropCrash").doubleValue() || !_bridgeadmin->isActive())) {
                delete (msg);
                _bridgeadmin->disableBridge();
                return;
            }

            // (2) Unterlassungsfehler
            if (rand() < RAND_MAX * par("failurePropOmission").doubleValue()) {
                delete (msg);
                _numOmission++;
                return;
            }
            // (3) Delay
            if (rand() < RAND_MAX * par("failurePropDelay").doubleValue()) {
                double delay = par("failureParamDelayDF").doubleValue();
                sendDelayed(msg, delay, "out");
                _numDelay++;
                return;
            }

            // (4) Datenfehler (Bitflip)
            if (rand() < RAND_MAX * par("failurePropBitflip").doubleValue()) {
                int numBitflips = par("failureParamBitflipNumber").intValue();
                FabMsg *fm = dynamic_cast<fabansimulator::FabMsg*>(msg);
                Data *dm = dynamic_cast<Data*>(fm->decapsulate());
                int bitflipPositions[numBitflips];
                for (int i = 0; i < numBitflips;) {
                    int pos = (int) ((((double) rand()) / (RAND_MAX - 1)) * (fm->getBitLength() + dm->getBitLength()));
                    bool unique = true;
                    for (int j = 0; j < i; j++) {
                        if (bitflipPositions[j] == pos)
                            unique = false;
                    }
                    if (unique) {
                        bitflipPositions[i] = pos;
                        if (pos < 16) {
                            uint16_t xormask = ((uint16_t) 1) << pos;
                            fm->setSenderID(fm->getSenderID() ^ xormask);
                        }
                        if (pos >= 16 && pos < 16 + 8) {
                            pos -= 16;
                            uint8_t xormask = ((uint8_t) 1) << pos;
                            fm->setSeqNo(fm->getSeqNo() ^ xormask);
                        }
                        if (pos >= 16 + 8 && pos < 16 + 8 + 64) {
                            pos -= 16 + 8;
                            uint64_t xormask = ((uint64_t) 1) << pos;
                            fm->setDeliveryTime((double) ((uint64_t) fm->getDeliveryTime() ^ xormask));
                        }
                        if (pos >= 16 + 8 + 64 && pos < 16 + 8 + 64 + 8) {
                            pos -= 16 + 8 + 64;
                            uint8_t xormask = ((uint8_t) 1) << pos;
                            fm->setHopCounter(fm->getHopCounter() ^ xormask);
                        }
                        if (pos >= 16 + 8 + 64 + 8 && pos < 16 + 8 + 64 + 8 + 16) {
                            pos -= 16 + 8 + 64 + 8;
                            uint8_t xormask = ((uint16_t) 1) << pos;
                            fm->setCheckingBridgeID(fm->getCheckingBridgeID() ^ xormask);
                        }
                        if (pos >= 16 + 8 + 64 + 8 + 16 && pos < 16 + 8 + 64 + 8 + 16 + 32) {
                            pos -= 16 + 8 + 64 + 8 + 16;
                            uint32_t xormask = ((uint32_t) 1) << pos;
                            fm->setSignature(fm->getSignature() ^ xormask);
                        }
                        if (pos >= 16 + 8 + 64 + 8 + 16 + 32) {
                            pos -= 16 + 8 + 64 + 8 + 16 + 32;
                            int dataArrElement = pos / 8;
                            pos = pos % 8;
                            dm->setData(dataArrElement, dm->getData(dataArrElement) ^ ((uint8_t) 1) << pos);
                        }
                        i++;
                    }
                }
                fm->encapsulate(dm);
                fm->setFaultinjections(fm->getFaultinjections() + 1);
                send(fm->dup(), "out");
                delete (msg);
                return;
            }

            // (4) Field Manipulation
            if (rand() < RAND_MAX * par("failurePropFieldManipulation").doubleValue()) {
                int numFields = par("failureParamFieldManipulationNumber").intValue();

                std::string s = par("failureParamFieldManipulationFields").stringValue();
                bool fields[6];

                int f = 0;
                for (int i = 0; i < 6; i++) {
                    if (s.at(i) == '1') {
                        f++;
                        fields[i] = true;
                    }
                    else {
                        fields[i] = false;
                    }
                }

                if (f < numFields) {
                    throw cRuntimeError("FieldManipulation parameters are incorrect. (%d < %d)", f, numFields);
                }

                FabMsg *fm = dynamic_cast<fabansimulator::FabMsg*>(msg);
                Data *dm = dynamic_cast<Data*>(fm->decapsulate());
                for (int i = 0; i < numFields;) {
                    int randField = (int) ((((double) rand()) / RAND_MAX) * 6);
                    if (fields[randField] == true) {
                        i++;
                        fields[randField] = false;

                        // Sender ID
                        if (randField == 0) {
                            int randSender = (int) ((_knownSenderIds.size() - 1) * ((double) rand()) / RAND_MAX);
                            fm->setSenderID(_knownSenderIds.at(randSender));
                        }
                        // Sequence Number
                        if (randField == 1) {
                            uint8_t seqNoShift = 1 + poisson(1, 0);
                            if (rand() < RAND_MAX / 2.0) {
                                fm->setSeqNo(fm->getSeqNo() + seqNoShift);
                            }
                            else {
                                fm->setSeqNo((uint8_t) std::max(0, (int) fm->getSeqNo() - (int) seqNoShift));
                            }
                        }
                        // Delivery Time
                        if (randField == 2) {
                            double factor = truncnormal(1, 2, 0);
                            fm->setDeliveryTime(fm->getDeliveryTime() * factor);
                        }
                        // Hop Counter
                        if (randField == 3) {
                            int hopShift = 1 + poisson(1, 0);
                            if (rand() < RAND_MAX / 2.0) {
                                fm->setHopCounter(fm->getHopCounter() + hopShift);
                            }
                            else {
                                fm->setHopCounter(std::max(0, fm->getHopCounter() - hopShift));
                            }
                        }

                        // Data
                        if (randField == 4) {
                            for (unsigned int i = 0; i < dm->getDataArraySize(); i++) {
                                dm->setData(i, (uint8_t) rand());
                            }
                        }
                        // Signature
                        if (randField >= 5) {
                            fm->setSignature((uint32_t) rand());
                        }
                    }
                }
                fm->encapsulate(dm);
                fm->setFaultinjections(fm->getFaultinjections() + 1);
                send(fm->dup(), "out");
                delete (msg);
                return;
            }

            // (5) Duplication
            if (rand() < RAND_MAX * par("failurePropDuplication").doubleValue()) {
                int dupNumber = par("failureParamDuplicationNumber").intValue();
                for (int i = 0; i < dupNumber; i++) {
                    send(fm->dup(), "out");
                }
                send(fm, "out");
                return;
            }

            // (7) Babbling Component
            if (rand() < RAND_MAX * par("failurePropBabblingComponent").doubleValue() && _bcCount == 0) {
                int bcNumber = par("failureParamBabblingComponentNumber").intValue();
                _bcCount += bcNumber;
                simtime_t t = simTime();

                for (int i = 0; i < bcNumber; i++) {
                    t += SimTime(par("failureParamBabblingComponentTimeInterval").doubleValue());
                    scheduleAt(t, new cMessage("BabblingComponentSelfMessage"));
                }
                fm->setFaultinjections(fm->getFaultinjections() + 1);
                send(fm, "out");
                return;
            }

            // (8)
            if (rand() < RAND_MAX * par("failurePropAddValidSignManipulation").doubleValue()) {
                int hop = fm->getHopCounter();
                double r = ((double) rand()) / RAND_MAX;
                double div = (par("faultassumption").intValue() > 1 ? 4.0 : 3.0);
                int sigFaultNumber = par("failureParamAddValidSignManipulationNumber").intValue();
                for (int i = 0; i < sigFaultNumber; i++) {
                    if (r < 1.0 / div) {
                        fm->setSignature(fm->getSignature() ^ _bridgeadmin->getNetworkAdmin()->getDmask());
                    }
                    else if (r >= 1.0 / div && r < 2.0 / div) {
                        fm->setSignature(fm->getSignature() ^ _bridgeadmin->getNetworkAdmin()->getCmask());
                    }
                    else if (r >= 2.0 / div && r <= 3.0 / div) {
                        uint32_t sig = fm->getSignature();
                        sig = rotateSignatureLeft(sig);
                        fm->setSignature(sig);
                    }
                    else {
                        fm->setSignature(fm->getSignature() ^ _bridgeadmin->getNetworkAdmin()->getPmask());
                    }
                }
                fm->setFaultinjections(fm->getFaultinjections() + 1);
                send(fm, "out");
                return;
            }

            send(msg, "out");
            return;

        }
        else {
            send(msg, "out");
        }
    }
    else {
        send(msg, "out");
    }
}

}
//namespace
