//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#include "routing.h"
#include "../../packets/fabMsg_m.h"

namespace fabansimulator
::bridge {

    Define_Module(Routing);

    void Routing::initialize()
    {
        _bridgeadmin = dynamic_cast<BridgeAdmin*>(this->getParentModule()->getSubmodule("bridgeAdmin"));
        _suppressedMessages = 0;
        inputStream.setName("InputStream");
        inputStream.record(0);
    }

    void Routing::_addAccHistory(uint16_t senderId, uint32_t sgn)
    {
        auto sid = accHistory.find(senderId);
        if(sid != accHistory.end()) {
            if(sid->second.size() < 5) {
                sid->second.push_back(sgn);
            }
            else {
                sid->second.erase(sid->second.begin());
                sid->second.push_back(sgn);
            }
        }
        else {
            std::vector<uint32_t> neuVec;
            neuVec.push_back(sgn);
            accHistory.insert(std::pair<uint16_t, std::vector<uint32_t>>(senderId, neuVec));
        }
    }

    bool Routing::_accHistoryContains(uint16_t senderId, uint32_t sgn) {
        auto sid = accHistory.find(senderId);
        if(sid!=accHistory.end()) {
            for (auto y = sid->second.begin(); y != sid->second.end(); y++) {
                if (sgn == *y) {
                    return true;
                }
            }
        }
        return false;
    }

    void Routing::_printHistory() {
        std::cout << "ACCHISTORY:" << std::endl;
        for (auto x : accHistory) {
            std::cout << x.first << " | ";
            for (auto y = x.second.begin(); y != x.second.end(); y++) {
                std::cout << *y << "\t";
            }
            std::cout << std::endl;
        }

    }

    void Routing::finish()
    {
        recordScalar("TimeCheckFailed", _suppressedMessages);
        recordScalar("InputStream", (_totalBits/simTime().dbl())/1000000.0,"MBit/s");

    }

    void Routing::refreshDisplay() const
    {
        char buf[40];
        sprintf(buf, "|del| = %d",_suppressedMessages);
        getDisplayString().setTagArg("t", 0, buf);
    }

    void Routing::handleMessage(cMessage *msg)
    {
        bool forw = false;
        if( dynamic_cast<FabMsg*>(msg)!= nullptr) {
            FabMsg* fm = dynamic_cast<FabMsg*>(msg);
            fm->setHops(fm->getHops()+1);

            int hop = fm->getHopCounter();

            _totalBits += fm->getBitLength();
            inputStream.record(_totalBits);
            _lastPacketReceivedTime = simTime();

            bool forwFailure=false;

            // Forwarding failure?
            if(_bridgeadmin->failactive()) {
                forwFailure = ((double)rand())/RAND_MAX < par("failurePropWrongForwarding").doubleValue();
            }

            if(_bridgeadmin->isDistributingBridge(fm)) {
                // Mode DB
                // Send to correct Bridges

                std::vector<int> rest;
                if(par("faultassumption").intValue()<=1) {
                    rest = _bridgeadmin->getEgrPortIndexListConnectedUnused(fm->getSenderID(), -1);
                }
                else {
                    rest = _bridgeadmin->getPrimaryEgrPortIndexListConnectedUnused(fm->getSenderID(), -1);
                }
                const std::vector<int> nodePorts = _bridgeadmin->getEgrPortIndexListNode();
                if(nodePorts.size()>0) {
                    rest.insert(end(rest), begin(nodePorts), end(nodePorts));
                }

                std::vector<int> egrPortList;
                if(par("faultassumption").intValue()<=1) {
                    egrPortList = _bridgeadmin->getPrimaryEgrPortIndexList(fm->getSenderID(), -1);
                }
                else {
                    egrPortList = _bridgeadmin->getPrimaryEgrPortIndexList(fm->getSenderID(), -1);
                }

                for(auto x : egrPortList) {
                    if(!forwFailure) {
                        send(msg->dup(), "out", x);
                    }
                    else if(((double)rand())/RAND_MAX < 0.5 ) {
                        send(msg->dup(), "out", x);
                    }
                    else {

                    }
                }

                if(forwFailure) {
                    for(auto x: rest) {
                        if(((double)rand())/RAND_MAX < 0.5 ) {
                            send(msg->dup(),"out", x );
                        }
                    }
                }
            }
            else if( _bridgeadmin->isPrecheckingBridge(fm)) {
                // Mode PB
                // Send to correct Bridges

                std::vector<int> rest;
                if(par("faultassumption").intValue()<=1) {
                    rest = _bridgeadmin->getEgrPortIndexListConnectedUnused(fm->getSenderID(), -1);
                }
                else {
                    rest = _bridgeadmin->getPrimaryEgrPortIndexListConnectedUnused(fm->getSenderID(), -1);
                }
                const std::vector<int> nodePorts = _bridgeadmin->getEgrPortIndexListNode();
                if(nodePorts.size()>0) {
                    rest.insert(end(rest), begin(nodePorts), end(nodePorts));
                }

                std::vector<int> egrPortList;
                if(par("faultassumption").intValue()<=1) {
                    egrPortList = _bridgeadmin->getEgrPortIndexList(fm->getSenderID(), fm->getRxPort());
                }
                else {
                    egrPortList = _bridgeadmin->getPrimaryEgrPortIndexList(fm->getSenderID(), fm->getRxPort());
                }

                for(auto x : egrPortList) {
                    if(!forwFailure) {
                        send(msg->dup(), "out", x);
                    }
                    else if(((double)rand())/RAND_MAX < 0.5 ) {
                        send(msg->dup(), "out", x);
                    }
                    else {

                    }
                }

                if(forwFailure) {
                    for(auto x: rest) {
                        if(((double)rand())/RAND_MAX < 0.5 ) {
                            send(msg->dup(),"out", x );
                        }
                    }
                }
            }
            else {
                // Mode CB only
                // DeliveryTime Verification
                bool cb = false;
                if(_bridgeadmin->isCheckingBridge(fm)) {
                    fm->setCheckingBridgeID(this->getParentModule()->getId());

                    if(par("dupDetectionRouting").boolValue() && fm->getFaultinjections()==0) {
                        Data* data = dynamic_cast<Data*>(fm->decapsulate());
                        uint8_t hopCounter = fm->getHopCounter();
                        fm->setHopCounter(0);
                        uint32_t crc = _bridgeadmin->generateCrc32(fm, data);
                        fm->setHopCounter(hopCounter);
                        fm->encapsulate(data);
                        if(_accHistoryContains(fm->getSenderID(), crc) && par("faultassumption").intValue()>1) {
                            delete(msg);
                            return;
                        }
                        else {
                            _addAccHistory(fm->getSenderID(), crc);
                        }
                    }

                    cb = true;
                    int h = _bridgeadmin->getNetworkAdmin()->getMaxWavelength(fm->getSenderID());
                    if(h!=-1) {
                        double deltaLink = (fm->getBitLength())/par("datarate").doubleValueInUnit("bps");
                        double deltaProc = (double)par("computingtime");
                        double deltaBusy = fm->getDeltaBusy();
                        double delTime = SimTime::fromRaw(fm->getDeliveryTime()).dbl();

                        if(delTime<0 || delTime >= omnetpp::SimTime::getMaxTime().dbl()) {
                            // Suppress Message
                            _suppressedMessages++;
                            delete(msg);
                            return;
                        }
                        simtime_t deliveryTime = SimTime::fromRaw(fm->getDeliveryTime());
                        simtime_t remainingTimeNeeded = par("deliveryTimeFactor").doubleValue() * SimTime((h * deltaLink + (h-1) * (deltaProc + deltaBusy)));
                        if(simTime() + remainingTimeNeeded > deliveryTime) {
                            // Suppress Message
                            _suppressedMessages++;
                            delete(msg);
                            return;
                        }
                    }

                    if(strcmp(fm->getCb(), "") == 0) {
                        fm->setCb(getParentModule()->getFullName());
                    }
                    else {
                    }
                }

                std::vector<int> rest = _bridgeadmin->getEgrPortIndexListConnectedUnused(fm->getSenderID(), -1);

                // Mode CB and FB
                // Send to correct bridges and all nodes
                for(auto x : _bridgeadmin->getEgrPortIndexList(_bridgeadmin->getNetworkAdmin()->getRoutingIdentifier(fm), (cb && par("faultassumption").intValue() >0 ? -1 : fm->getRxPort()))) {
                    if(!forwFailure) {
                        send(msg->dup(), "out", x);
                    }
                    else if(((double)rand())/RAND_MAX < 0.5 ) {
                        send(msg->dup(), "out", x);
                    }
                    else {

                    }
                }

                if(forwFailure) {
                    for(auto x: rest) {
                        if(((double)rand())/RAND_MAX < 0.5 ) {
                            send(msg->dup(),"out", x );
                        }
                    }
                }

            }
            delete(msg);

        }
    }

}
//namespace
