//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#include "queue.h"
#include "../packets/fabMsg_m.h"

namespace fabansimulator {

Define_Module(Queue);

void Queue::initialize() {
    _queueCapacity = par("queuesize");
    if (gate("out")->getNextGate()->getNextGate() != nullptr)
        _outChannel = gate("out")->getTransmissionChannel();
    else
        _outChannel = nullptr;

    // Initialize Statistics
    _queueLoadPackets.setName("QueueLoad [packets]");
    _queueLoadByte.setName("QueueLoad [Byte]");
    _queueLoadPackets.record(0);
    _queueLoadByte.record(0);
}

void Queue::finish() {
    _queueLoadPackets.record(_queue.getBitLength());
    _queueLoadByte.record(_queue.getByteLength());
    _queueingTime.recordAs("QueueingTime", "s");
}

void Queue::refreshDisplay() const {
    char buf[40];
    sprintf(buf, "|Q| = %dP / %ldbyte", _queue.getLength(), _queue.getByteLength());
    getDisplayString().setTagArg("t", 0, buf);
}

void Queue::handleMessage(cMessage *msg) {
    if (dynamic_cast<FabMsg*>(msg) != nullptr) {
        if (_outChannel->isBusy()) {
            if (_queue.getByteLength() + dynamic_cast<FabMsg*>(msg)->getByteLength() <= _queueCapacity) {
                msg->setTimestamp();
                _queue.insert((cPacket*) msg);

                // Update Statistics
                _queueLoadPackets.record(_queue.getLength());
                _queueLoadByte.record(_queue.getByteLength());
            }
            else {
                delete msg;
                EV_FATAL << "Message dropped in " << this->getFullPath() << std::endl;
            }

            if (_queue.getLength() == 1) {
                scheduleAt(_outChannel->getTransmissionFinishTime(), new cMessage("popQueue"));
            }
        }
        else {
            send(msg, "out");
        }
    }
    else if (strcmp(msg->getName(), "popQueue") == 0) {
        if (_queue.getLength() == 0)
            throw cRuntimeError("Trying to pop message though queue is empty (%s)",
                    this->getParentModule()->getFullName());
        cMessage* m = _queue.pop();
        _queueingTime.collect(simTime() - m->getTimestamp());
        FabMsg* fm = dynamic_cast<FabMsg*>(m);
        fm->getBusyVector().push_back(std::pair<std::string,double>(this->getFullPath(),(simTime() - m->getTimestamp()).dbl()));
        send(m, "out");

        // Update Statistics
        _queueLoadPackets.record(_queue.getLength());
        _queueLoadByte.record(_queue.getByteLength());

        if (_queue.getLength() > 0) {
            scheduleAt(_outChannel->getTransmissionFinishTime(), msg);
        }
        else {
            delete msg;
        }
    }
}

}
//namespace
