//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#include <fstream>
#include "../networkadmin/networkadmin.h"
#include "../bridge/modules/bridgeAdmin.h"
#include "../node/modules/nodeAdmin.h"
#include "../FabanSimLib.h"

namespace fabansimulator {

Define_Module(NetworkAdmin);

void NetworkAdmin::initialize() {

    // Retrieve list of all nodes in the network
    for (cModule::SubmoduleIterator it(this->getParentModule()); !it.end(); it++) {
        cModule *submodule = *it;
        if (strcmp(submodule->getModuleType()->getName(), "Node") == 0) {
            cModule *node = submodule;
            if (node == nullptr) {
                throw cRuntimeError("Node %s does not contain a \"nodeadmin\" submodule!", submodule->getName());
            }
            _nodelist.push_back(node);
        }
        if (strcmp(submodule->getModuleType()->getName(), "Bridge") == 0) {
            cModule* bridge = submodule;
            _bridgelist.push_back(bridge);
        }

    }

    EV << "Number of Bridges: " << _bridgelist.size() << endl;
    EV << "Number of Nodes:   " << _nodelist.size() << endl;

    // Explore network and create json file for FabanWaveTool
    _topo.extractByNedTypeName(cStringTokenizer("fabansimulator.bridge.Bridge fabansimulator.node.Node").asVector());

    json js;
    js["topology"] = {};
    js["topology"]["endstations"] = {};
    js["topology"]["bridges"] = {};
    js["topology"]["links"] = {};

    for (int i = 0; i < _topo.getNumNodes(); i++) {
        cTopology::Node *node = _topo.getNode(i);
        if (strcmp(node->getModule()->getModuleType()->getName(), "Node") == 0) {
            js["topology"]["endstations"].push_back(node->getModule()->getFullName());
        }
        if (strcmp(node->getModule()->getModuleType()->getName(), "Bridge") == 0) {
            js["topology"]["bridges"].push_back(node->getModule()->getFullName());
        }
        for (int j = 0; j < node->getNumOutLinks(); j++) {
            cTopology::LinkOut* link = node->getLinkOut(j);
            json l, l1, l2;
            std::string nameA = link->getLocalNode()->getModule()->getFullName();
            std::string nameB = link->getRemoteNode()->getModule()->getFullName();
            if (strcmp(nameA.c_str(), nameB.c_str()) < 1) {
                l1 = {"A", { {"name", nameA}, {"gate", link->getLocalGate()->getIndex()}}};
                l2 = {"B", { {"name", nameB}, {"gate", link->getRemoteGate()->getIndex()}}};
                l = {l1,l2};
                js["topology"]["links"].push_back(l);
                EV << node->getModule()->getName() << " --> " << link->getRemoteNode()->getModule()->getName() << endl;
            }
        }
    }

    if (strcmp(par("routing").stringValue(), "auto") == 0) {
        std::cout << "INFO: Routing will be generated" << std::endl;
        std::ofstream file("top.json");
        file << js.dump(4);
        if (system("java -jar ../ext/Fabanwavetool.jar -i top.json -create -O") != 0) {
            throw cRuntimeError("FabanWaveTool call not successful.");
        }
    }

    // Retrieve all connections between bridges and nodes and inform corresponding brdiges about this information
        for (int i = 0; i < _topo.getNumNodes(); i++) {
            if (strcmp(_topo.getNode(i)->getModule()->getModuleType()->getName(), "Node") == 0) {
                for (int j = 0; j < _topo.getNode(i)->getNumInLinks(); j++) {
                    if (strcmp(_topo.getNode(i)->getLinkIn(j)->getRemoteNode()->getModule()->getModuleType()->getName(), "Bridge") == 0) {
                        dynamic_cast<bridge::BridgeAdmin*>(_topo.getNode(i)->getLinkIn(j)->getRemoteNode()->getModule()->getSubmodule("bridgeAdmin"))->addPortToNodeInformation(_topo.getNode(i)->getLinkIn(j)->getRemoteGate()->getIndex());
                    }
                }
            }
        }

    // Read FabanConfiguration
    std::string filename = "top.json";
    if (strcmp(par("routing").stringValue(), "auto") != 0) {
        filename = par("routing").stringValue();
        std::cout << "INFO: Routing file \"" << filename << "\" used" << std::endl;
    }
    std::ifstream ifs(filename);
    json js2 = json::parse(ifs);

    if (!js2.contains("fabanconfiguration")) {
        cRuntimeError("No Faban Configuration Found or possible to create");
    }
    json::array_t config = js2["fabanconfiguration"];
    if (par("faultassumption").intValue() == 1) {
        std::cout << "INFO: Faultassumption = 1   ==>   Legacy Routing format reading" << std::endl;
        for (int i = 0; i < config.size(); i++) {
            std::string db = config.at(i)["distributingbridge"];
            int wavelength = 0;
            for (auto it = config.at(i).begin(); it != config.at(i).end(); it++) {
                if (it.key().find("wave", 0) != std::string::npos) {
                    _linkBlacklist.clear();
                    std::string cb = config.at(i)[it.key()]["checkingbridge"];
                    _addCheckingBridgeToMap(db, cb);
//                dynamic_cast<bridge::BridgeAdmin*>(_getModuleFromNode(db)->getSubmodule("bridgeAdmin"))->addPrimaryRoutingInformation(db, -1, _getOutPort("", db, cb));
                    dynamic_cast<bridge::BridgeAdmin*>(_getModuleFromNode(db)->getSubmodule("bridgeAdmin"))->addPrimaryRoutingInformation(db, -1, _getOutPort("", db, cb));
                    _linkBlacklist.push_back(db);


                    _recursivePrimaryRouting(config.at(i)[it.key()]["primaryrouting"], db + "." + cb, db, config.at(i)[it.key()]["checkingbridge"]);
                    _linkBlacklist.clear();
                    _linkBlacklist.push_back(db);
                    dynamic_cast<bridge::BridgeAdmin*>(_getModuleFromNode(cb)->getSubmodule("bridgeAdmin"))->addRoutingInformationNode(db + "." + cb, -1);
                    wavelength = std::max(wavelength, 1 + _recursiveRouting(config.at(i)[it.key()]["routing"], db + "." + cb, db, config.at(i)[it.key()]["checkingbridge"]));

                }
            }
            _dbWavelengthMap.insert(std::pair<std::string, int>(db, wavelength));
        }
    }
    else {
        std::cout << "INFO: Faultassumption > 1   ==>   New P/S Routing format reading" << std::endl;
        for (int i = 0; i < config.size(); i++) {
            std::cout<< "Itertation: " << i << std::endl;
            auto fabanconfigIt = config.at(i);

            std::string db = fabanconfigIt["distributingbridge"];
            std::cout<< "Itertation DB: " << db << std::endl;
            int maxWaveLength = 0;
            for (auto waveconfigIt = fabanconfigIt.begin(); waveconfigIt != fabanconfigIt.end(); waveconfigIt++) {
                if (waveconfigIt.key().find("wave", 0) != std::string::npos) {
                    std::cout<< "Wave found: " << waveconfigIt.key() << std::endl;
                    std::string cb = fabanconfigIt[waveconfigIt.key()]["checkingbridge"];
                    std::cout<< "CB: " <<cb << std::endl;
                    _addCheckingBridgeToMap(db, cb);
//                    std::cout << db << " -> ";
                    std::cout<< "Added to map" << std::endl;
                    std::cout << std::endl;
                    int primMax = 0;
                    int secMax = 0;
                    std::cout<< "" << std::endl;
                    for (auto dbNeighbour : waveconfigIt.value()["primaryrouting"][db]) {
                        std::cout<< "For loop entered" << std::endl;
//                        std::cout << dbNeighbour << " ";
                        dynamic_cast<bridge::BridgeAdmin*>(_getModuleFromNode(db)->getSubmodule("bridgeAdmin"))->addPrimaryRoutingInformation(db, -1, _getOutPort("", db, dbNeighbour));
                        std::cout<< "casted" << std::endl;
                        int x = 1 + _recursivePrimaryRouting(fabanconfigIt[waveconfigIt.key()]["primaryrouting"], db, db, dbNeighbour);
                        primMax = std::max(primMax, x);
                        std::cout << "[" << db << "->" << dbNeighbour << "] Prim Routing Length = " << x << std::endl;
                    }
                    std::cout<< "For Loop 1 finished "<< std::endl;
//                    std::cout << cb << " ----> ";

                    for (auto rout = fabanconfigIt[waveconfigIt.key()].begin(); rout != fabanconfigIt[waveconfigIt.key()].end(); rout++) {
                        if (rout.key().find("routing", 0) != std::string::npos && rout.key().find("primaryrouting", 0) == std::string::npos) {
                            std::cout << "Routing found: " << rout.key() << std::endl;

                            _linkBlacklist.clear();
                            for (auto cbNeighbour : waveconfigIt.value()[rout.key()][cb]) {
//                                std::cout << cbNeighbour << " ";
                                dynamic_cast<bridge::BridgeAdmin*>(_getModuleFromNode(cb)->getSubmodule("bridgeAdmin"))->addRoutingInformation(db + "." + cb, -1, _getOutPort("", cb, cbNeighbour));
                                dynamic_cast<bridge::BridgeAdmin*>(_getModuleFromNode(cb)->getSubmodule("bridgeAdmin"))->addRoutingInformationNode(db + "." + cb, -1);
                                int x = 1 + _recursiveRouting(fabanconfigIt[waveconfigIt.key()][rout.key()], db + "." + cb, cb, cbNeighbour);
                                secMax = std::max(secMax, x);
                                std::cout << "[" << db << "." << cb << "." << cbNeighbour << "] Prim Routing Length = " << x << std::endl;
                            }
                            std::cout<< "Inner Loop iterated"<< std::endl;
                        }
                    }
                    std::cout<< "For Loop 2 finished "<< std::endl;
                    maxWaveLength = std::max(maxWaveLength, primMax + secMax);

//                    std::cout << std::endl;

                }
            }

            std::cout << "\n\n===============================================" << std::endl;
            std::cout << "[" << db << "] MaxWaveLength = " << maxWaveLength << std::endl;
            std::cout << "===============================================" << std::endl << std::endl;
            _dbWavelengthMap.insert(std::pair<std::string, int>(db, maxWaveLength));
        }
    }

    // Generate signature masks
    d_mask = (uint32_t) rand();
    c_mask = (uint32_t) rand();
    p_mask = (uint32_t) rand();
    int f = par("faultassumption");

    r_mask = rotateSignatureRight(c_mask, f - 1 + 2);
    std::cout << std::hex << r_mask << " --> ";
    for (int j = 2; j <= f - 1 + 1; j++) {
        r_mask = r_mask ^ rotateSignatureRight(p_mask, j);
        std::cout << r_mask << " --> ";
    }
    r_mask = r_mask ^ rotateSignatureRight(d_mask, 1);
    std::cout << std::hex << r_mask << " = RMASK " << std::endl;

    std::cout << std::hex << std::uppercase << "DMASK: " << d_mask << "\tPMASK: " << p_mask << "\tCMASK: " << c_mask << "\tRMASK: " << r_mask << std::endl;

//    _printCheckingBridgeMap();
}

void NetworkAdmin::finish() {
    recordScalar("SimDuration", simTime());

    // Remove Topology Json File From Filesystem if it was auto-generated
    if (strcmp(par("routing").stringValue(), "auto") == 0) {
        std::cout << "INFO: Routing file deleted" << std::endl;
        remove("top.json");
    }
}

void NetworkAdmin::distributePublicKeyPair(uint16_t id, uint32_t q, uint32_t r) {
    for (std::vector<cModule*>::iterator it = _nodelist.begin(); it != _nodelist.end(); it++) {
        node::NodeAdmin* nodeadmin = dynamic_cast<node::NodeAdmin*>((*it)->getSubmodule("nodeadmin"));
        nodeadmin->addPublicSignatureKeys(id, q, r);
    }
}

std::string NetworkAdmin::getDistributingBridgeName(uint16_t senderNode) {
    if (_senderIdBridgeMap.size() == 0) {
        // Initialize Map
        for (int i = 0; i < _topo.getNumNodes(); i++) {
            if (strcmp(_topo.getNode(i)->getModule()->getModuleType()->getName(), "Node") == 0) {
                for (int j = 0; j < _topo.getNode(i)->getNumOutLinks(); j++) {
                    if (strcmp(_topo.getNode(i)->getLinkOut(j)->getRemoteNode()->getModule()->getModuleType()->getName(), "Bridge") == 0) {
                        std::string db = _topo.getNode(i)->getLinkOut(j)->getRemoteNode()->getModule()->getFullName();
                        uint16_t n = dynamic_cast<node::NodeAdmin*>(_topo.getNode(i)->getModule()->getSubmodule("nodeadmin"))->getId();
                        _senderIdBridgeMap.insert(std::pair<uint16_t, std::string>(n, db));
                    }
                }
            }
        }
    }

    if (_senderIdBridgeMap.find(senderNode) == _senderIdBridgeMap.end())
        return "";

    return _senderIdBridgeMap.find(senderNode)->second;
}

void NetworkAdmin::handleMessage(cMessage *msg) {

}

cModule* NetworkAdmin::_getModuleFromNode(std::string name) {

    int posBrack = name.find("[", 0);
    if (posBrack > 0) {
        return this->getParentModule()->getSubmodule(name.substr(0, name.find("[", 0)).c_str(), atoi(name.substr(posBrack + 1, name.find("]", 0) - posBrack - 1).c_str()));
    }
    else {
        return this->getParentModule()->getSubmodule(name.c_str());
    }
}

int NetworkAdmin::getNumberOfBridges() {
    return _bridgelist.size();
}

int NetworkAdmin::getMaxDataSize() {
    return par("maxDataSize").intValue();
}

int NetworkAdmin::_recursiveRouting(json routing, std::string db, std::string prevNode, std::string currNode) {
    std::stringstream ss;
    ss << currNode;
    std::cout << db << " -> " << prevNode << " - " << currNode<< std::endl;

    dynamic_cast<bridge::BridgeAdmin*>(_getModuleFromNode(currNode)->getSubmodule("bridgeAdmin"))->addRoutingInformationNode(db, _getInPort(prevNode, currNode, ""));
//    dynamic_cast<bridge::BridgeAdmin*>(_getModuleFromNode(currNode)->getSubmodule("bridgeAdmin"))->addRoutingInformation

    if (std::find(_linkBlacklist.begin(), _linkBlacklist.end(), ss.str().c_str()) != _linkBlacklist.end()) {
        std::cerr << "Aborted!" << std::endl;
        return 0;
    }

    _linkBlacklist.push_back(ss.str());

//    std::cout << db << " -> " << currNode<< std::endl;
    int wltemp = 0;
    for (int i = 0; i < routing[currNode].size(); i++) {
        int portIn = _getInPort(prevNode, currNode, routing[currNode].at(i));
        int portOut = _getOutPort(prevNode, currNode, routing[currNode].at(i));
        dynamic_cast<bridge::BridgeAdmin*>(_getModuleFromNode(currNode)->getSubmodule("bridgeAdmin"))->addRoutingInformation(db, (db.compare(prevNode + "." + currNode) == 0 ? -1:portIn), portOut);
        wltemp = std::max(1 + _recursiveRouting(routing, db, currNode, routing[currNode].at(i)), wltemp);
    }
    return wltemp;

}

int NetworkAdmin::_recursivePrimaryRouting(json routing, std::string db, std::string prevNode, std::string currNode) {
    std::stringstream ss;
    ss << currNode;
//    std::cout<< "Rec Prim: "<< db << " -> " << prevNode << " -> " << currNode << std::endl;

    int wltemp = 0;
    for (int i = 0; i < routing[currNode].size(); i++) {
        int portIn = _getInPort(prevNode, currNode, routing[currNode].at(i));
        int portOut = _getOutPort(prevNode, currNode, routing[currNode].at(i));
        dynamic_cast<bridge::BridgeAdmin*>(_getModuleFromNode(currNode)->getSubmodule("bridgeAdmin"))->addPrimaryRoutingInformation(db, portIn, portOut);
        wltemp = std::max(1 + _recursivePrimaryRouting(routing, db, currNode, routing[currNode].at(i)), wltemp);
    }
    return wltemp;

}

int NetworkAdmin::_getInPort(std::string prevNode, std::string node, std::string nextNode) {
    cTopology::Node* pN = _topo.getNodeFor(_getModuleFromNode(prevNode));
    for (int i = 0; i < pN->getNumOutLinks(); i++) {
        cTopology::LinkOut* link = pN->getLinkOut(i);
        if (strcmp(link->getRemoteNode()->getModule()->getFullName(), node.c_str()) == 0) {
            return link->getRemoteGate()->getIndex();
        }
    }
    return -1;
}

uint32_t NetworkAdmin::getDmask() {
    return d_mask;
}

uint32_t NetworkAdmin::getCmask() {
    return c_mask;
}

uint32_t NetworkAdmin::getPmask() {
    return p_mask;
}

uint32_t NetworkAdmin::getRmask() {
    return r_mask;
}

int NetworkAdmin::getMaxWavelength(uint16_t senderID) {
    std::string db = getDistributingBridgeName(senderID);
    std::map<std::string, int>::iterator element = _dbWavelengthMap.find(db);
    if (element == _dbWavelengthMap.end()) {
        return -1;
    }
    return element->second;
}

int NetworkAdmin::_getOutPort(std::string prevNode, std::string node, std::string nextNode) {
    cTopology::Node* n = _topo.getNodeFor(_getModuleFromNode(node));
    for (int i = 0; i < n->getNumOutLinks(); i++) {
        cTopology::LinkOut* link = n->getLinkOut(i);
        if (strcmp(link->getRemoteNode()->getModule()->getFullName(), nextNode.c_str()) == 0) {
            return link->getLocalGate()->getIndex();
        }
    }
    return -1;
}

void NetworkAdmin::_addCheckingBridgeToMap(std::string db, std::string cb) {
    auto it = _checkingBridgeMap.find(db);
    if (it != _checkingBridgeMap.end()) {
        it->second.push_back(cb);
    }
    else {
        std::vector<std::string> newvector;
        newvector.push_back(cb);
        _checkingBridgeMap.insert(std::pair<std::string, std::vector<std::string>>(db, newvector));
    }
}

bool NetworkAdmin::isCheckingBridge(std::string db, std::string cb) {
    auto it = _checkingBridgeMap.find(db);
    if (it == _checkingBridgeMap.end())
        return false;
    for (std::string c : it->second) {
        if (c.compare(cb) == 0)
            return true;
    }
    return false;
}

std::string NetworkAdmin::getRoutingIdentifier(FabMsg* fm) {
//    if (par("faultassumption").intValue() > 1) {
        return getDistributingBridgeName(fm->getSenderID()) + "." + fm->getCb();
//    }
//    else {
//        return getDistributingBridgeName(fm->getSenderID());
//    }
}

void NetworkAdmin::_printCheckingBridgeMap() {
    std::cout << "\n==================================\n CHECKING BRIDGE MAP \n==================================" << std::endl;
    for (auto db : _checkingBridgeMap) {
        std::cout << "\t" << db.first << ": ";
        for (auto cb : db.second) {
            std::cout << cb << " ";
        }
        std::cout << std::endl;
    }
    std::cout << "==================================" << std::endl;
}

} //namespace

