//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#include "sink.h"
#include "../../packets/fabMsg_m.h"
#include "../../packets/data_m.h"
#include "../../FabanSimLib.h"

namespace fabansimulator
::node {

    Define_Module(Sink);

    void Sink::initialize()
    {
        _numRx = 0;
        _nodeadmin = (NodeAdmin*)(this->getParentModule()->getSubmodule("nodeadmin"));
        _numDiscardedDelay = 0;
        _numDiscardedFaulty = 0;
        _numSuccessfulBc = 0;
        _numFfDuplicates = 0;
        _numSuccessfulBcFromFaultyBridges = 0;
    }

    void Sink::refreshDisplay() const
    {
        char buf[40];
        sprintf(buf, "RX: %ld", _numRx);
        getDisplayString().setTagArg("t", 0, buf);
    }

    void Sink::handleMessage(cMessage *msg)
    {
        if(dynamic_cast<FabMsg*>(msg) != nullptr) {
            if(!msg->isSelfMessage()) {
                _numRx++;
                FabMsg* fm = dynamic_cast<FabMsg*>(msg);
                if(fm->getDeliveryTime()<0 || (SimTime::fromRaw(fm->getDeliveryTime()))>= omnetpp::SimTime::getMaxTime()) {
                    // Invalid Delivery Time
                    std::cout << "YES" << std::endl;
                    _numDiscardedFaulty++;
                    delete(fm);
                }
                else {
                    _diffToTd.collect(SimTime::fromRaw(fm->getDeliveryTime()) - simTime());
                    simtime_t delTime = SimTime::fromRaw(fm->getDeliveryTime());
                    _rxTimes.collect(msg->getArrivalTime() - msg->getCreationTime());
                    _rxTimesNormalized.collect((msg->getArrivalTime() - msg->getCreationTime())/(fm->getByteLength()));
                    if(delTime < simTime()) {
                        // Message is too late
                        _numDiscardedDelay++;
                        delete(fm);
                    }
                    else {
                        scheduleAt(delTime, msg);
                    }
                }
            }
            else {
                FabMsg* fm = dynamic_cast<FabMsg*>(msg);
                Data* data = dynamic_cast<Data*>(fm->decapsulate());

                uint32_t idSgn = fm->getSignature();

                // Signature modification
                fm->setHopCounter(0);
                fm->setSignature(rotateSignatureRight(fm->getSignature(), 1+(int)par("faultassumption"))^_nodeadmin->getNetworkAdmin()->getRmask());



                // Verfiy message integrity
                std::pair<uint32_t, uint32_t> keyPairQR = _nodeadmin->getPublicKeys(fm->getSenderID());
                uint32_t crc = _nodeadmin->generateCrc32(fm, data);
                bool sgnResult = validationOddMulSignature(crc, keyPairQR.second, keyPairQR.first, fm->getSignature());
                if(!sgnResult) {
                    // Message is faulty
                    EV_ERROR << "Signature test failed. Corrupted Message received." << std::endl;
                    _numDiscardedFaulty++;
                }
                else {

                    if(_accHistoryContains(fm->getSenderID(), idSgn)) {
                        _numFfDuplicates++;
                    }
                    else {
                        _addAccHistory(fm->getSenderID(), idSgn);
                        if(par("sepFaultyBridgeStats").str().find(_nodeadmin->getNetworkAdmin()->getDistributingBridgeName(fm->getSenderID()),0)==std::string::npos){
                            _numSuccessfulBc++;
                        }else{
                            _numSuccessfulBcFromFaultyBridges++;
                        }

                        _numRxSender.collect(fm->getSenderID());
                    }

                }
                delete(data);
                delete(fm);
            }
        }
        else {
            EV_ERROR << "Unknown message received!" << std::endl;
            delete(msg);
        }

    }

    void Sink::finish()
    {
        recordScalar("NumRX", _numRx);
        recordScalar("NumDiscardedDelay", _numDiscardedDelay);
        recordScalar("NumDiscardedFaulty", _numDiscardedFaulty);
        recordScalar("NumFfDuplicates", _numFfDuplicates);
        recordScalar("NumSuccessfulBroadcasts", _numSuccessfulBc);
        recordScalar("NumSuccessfulBroadcastsFromFaultyBridges", _numSuccessfulBcFromFaultyBridges);
        _rxTimes.recordAs("RxTimes");
        _rxTimesNormalized.recordAs("RxTimes Normalized");
        _diffToTd.recordAs("DifferenceToDeliveryDeadline");
        _numRxSender.recordAs("NumRxFromSender");
    }

    void Sink::_addAccHistory(uint16_t senderId, uint32_t sgn)
    {
        auto sid = accHistory.find(senderId);
        if(sid != accHistory.end()) {
            if(sid->second.size() < 5) {
                sid->second.push_back(sgn);
            }
            else {
                sid->second.erase(sid->second.begin());
                sid->second.push_back(sgn);
            }
        }
        else {
            std::vector<uint32_t> neuVec;
            neuVec.push_back(sgn);
            accHistory.insert(std::pair<uint16_t, std::vector<uint32_t>>(senderId, neuVec));
        }
    }

    bool Sink::_accHistoryContains(uint16_t senderId, uint32_t sgn) {
        auto sid = accHistory.find(senderId);
        if(sid!=accHistory.end()) {
            for (auto y = sid->second.begin(); y != sid->second.end(); y++) {
                if (sgn == *y) {
                    return true;
                }
            }
        }
        return false;
    }

    void Sink::_printHistory() {
        std::cout << "ACCHISTORY:" << std::endl;
        for (auto x : accHistory) {
            std::cout << x.first << " | ";
            for (auto y = x.second.begin(); y != x.second.end(); y++) {
                std::cout << *y << "\t";
            }
            std::cout << std::endl;
        }

    }

}

//namespace
