#include "source.h"

#include "../../FabanSimLib.h"
#include "../../networkadmin/networkadmin.h"

namespace fabansimulator::node {

Define_Module(Source);
volatile double nextSchedule;

void Source::initialize() {
    _numTx = 0;

    if (par("sendInterval").doubleValue() <= 0 && par("sendRate").doubleValue() <= 0) {
        throw cRuntimeError("Neither sendInterval nor sendRate is set. No send behavior is specified.");
    }
    if (par("sendInterval").doubleValue() > 0 && par("sendRate").doubleValue() > 0) {
        throw cRuntimeError("Contradictory parameters (sendInterval and sendRate) are set. Specify only one of them!");
    }
    if (par("sendRate").doubleValue() > 0) {
        _txmode = TXMODE_RATE;
    }
    else {
        _txmode = TXMODE_INTERVAL;
    }

    active = par("srcGeneratorActive");
    if (active && par("maxMessages").intValue() != 0) {
        _firstMsgSent = simTime() + _nextTxTimePoint() + par("timeShift");
        scheduleAt(_firstMsgSent, new cMessage("genNewFabanMsg"));
    }

    // Retrieve NodeAdmin Pointer
    cModule *nodeAdminCmod = this->getParentModule()->getSubmodule("nodeadmin");
    if (nodeAdminCmod == nullptr) {
        throw cRuntimeError("No Local Module named \"nodeadmin\" inside Node found!");
    }
    nodeAdmin = dynamic_cast<NodeAdmin*>(nodeAdminCmod);

    // Retrieve maxDataSize
    _maxDataSize = nodeAdmin->getNetworkAdmin()->getMaxDataSize();

    // Calculate DeltaBusy

    _deltaBusy = par("deltaBusy").doubleValueInUnit("s");
    if (_deltaBusy >= 0) {
        std::cout << "INFO: DeltaBusy set via configuration to " << _deltaBusy << std::endl;
    }
}

void Source::handleMessage(cMessage *msg) {
    if (strcmp(msg->getName(), "genNewFabanMsg") == 0) {
        _numTx++;

        // Create new Faban Packet
        FabMsg *fm = new FabMsg(_newFabanMsgName().c_str(), 0);
        fm->setSenderID(nodeAdmin->getId());
        fm->setSeqNo(nodeAdmin->getSeqNo());
        nodeAdmin->incSeqNo();

        // Create Data Packet
        Data *data = new Data("", 0);
        _lastDatasize = par("dataSize").intValue();
        if (_lastDatasize > _maxDataSize) {
            _lastDatasize = _maxDataSize;
        }
        data->setDataArraySize(_lastDatasize);
        data->setByteLength(_lastDatasize);
        //            _createRandomData(data);

        // Set remaining fields
        fm->setHopCounter(0);
        fm->setBitLength(fm->getPacketSizeInBits());

        //Calculate DeliveryTime
        _maxHops = nodeAdmin->getNetworkAdmin()->getMaxWavelength(nodeAdmin->getId());
        if (_deltaBusy == -1) {
            _deltaBusy = par("deltaBusyFactor").doubleValue() * (2 * nodeAdmin->getNetworkAdmin()->getNumberOfBridges() - 1) * 8 * (_maxDataSize + 16) / par("datarate").doubleValueInUnit("bps") / (_maxHops + 1);
            std::cout << "INFO: DeltaBusy calculated to " << _deltaBusy << std::endl;
        }
        //            double deltaLink = (fm->getBitLength() + data->getBitLength())/(par("datarate").doubleValueInUnit("bps")/par("noPorts").intValue());
        double deltaLink = (fm->getBitLength() + data->getBitLength()) / (par("datarate").doubleValueInUnit("bps"));
        double deltaProc = par("computingtime").doubleValueInUnit("s");

        double deliveryTime = par("deliveryTimeFactor").doubleValue() * ((_maxHops + 2) * deltaLink + (_maxHops + 1) * (deltaProc + _deltaBusy));
        simtime_t delTime = simTime() + SimTime(deliveryTime);
        fm->setDeliveryTime(delTime.raw());
        fm->setDeltaBusy(_deltaBusy);

        // Calculate CRC + MsigOM
        fm->setSignature(calculateOddMulSignature(nodeAdmin->generateCrc32(fm, data), nodeAdmin->getSgnKeyP()));

        // Encapsulate Data Packet and send the Faban Packet
        fm->encapsulate(data);
        _totalBitSent += fm->getBitLength();
        send(fm, "out");

        // Schedule next message generation
        if (_numTx < par("maxMessages").intValue() || par("maxMessages").intValue() == -1) {
            nextSchedule = _nextTxTimePoint() + par("timeShift").doubleValueInUnit("s");
            scheduleAt(simTime() + nextSchedule, msg);
        }
        else {
            delete msg;
        }
    }
}

void Source::finish() {
    recordScalar("NumTX", _numTx);
    recordScalar("RateTX", _datarate, "Mbps");
}

std::string Source::_newFabanMsgName() {
    std::stringstream ss;
    ss << this->getParentModule()->getFullName() << "." << (int) nodeAdmin->getSeqNo();
    return ss.str();
}

void Source::refreshDisplay() const {
    char buf[40];
    sprintf(buf, "TX: %ld", _numTx);
    getDisplayString().setTagArg("t", 0, buf);
}

void Source::_createRandomData(Data *dataMsg) {
    for (unsigned int i = 0; i < dataMsg->getDataArraySize(); i++) {
        dataMsg->setData(i, (uint8_t) rand());
    }
}

double Source::_nextTxTimePoint() {
    switch (_txmode) {
        case TXMODE_INTERVAL:
            _datarate = _totalBitSent / (1000000.0 * (simTime().dbl()));
            return (par("sendInterval").doubleValue());
            break;
        case TXMODE_RATE:
            if (_lastDatasize <= 0) {
                return 0;
            }
            else {
                double nextDelta = (8 * (18 + _lastDatasize)) / par("sendRate").doubleValueInUnit("bps");
                _datarate = _totalBitSent / (1000000.0 * (simTime().dbl() + nextDelta));
                return nextDelta;
            }
            return -1;
            break;
        default:
            throw cRuntimeError("_txmode not set correctly!");
            break;
    }
}

}
//namespace
