/*
 * GlobalBpiLib.h
 *
 *      Author: Valentin Fitz
 *   Institute: University of Duisburg-Essen, Germany
 */

/**
 * @file
 *
 * Provides a number of general functions and classes for different use cases, which should be needed in all
 * implementations relating the BananaNetworkTower Project. This class needs obligatory the counterpart header
 * LocalBpiLib.h, containing the following definitions
 * 	- NAMESTAMP
 *
 * */
#ifndef GLOBALBPILIB_HPP_
#define GLOBALBPILIB_HPP_

//#include "LocalBpiLib.h"
#include "include/json.hpp"
#include <string>
#include <sstream>
#include <vector>
#include <csignal>
#include <chrono>
#include <time.h>

#define STAMPLENGTH 6 /**< Length of Namestamps. */
#define GEN_BUFFERSIZE 4096
#define DLE "\x10"
/*FIXME JN*/
#define TEN_POWER_NINE 1000000000
#define JSON_CONFIGFILE "/home/user/Banananetwork/.config/ioconfig.json"
/*ENd JN*/

//#define ST_UDP 1
//#define ST_RAW 0
//#define SOCKETTYPE ST_RAW

/**
 * \brief Custom bpi printer class
 *
 * Implementation of an ostream printer with preceding namestamps. This provides a consisten output for
 * different applications.
 */
class BpiOutStream {
private:
	std::ostream& _os; /**< std::ostream which prints the output. */
	std::stringstream _localStringStream; /**< Storage for local and temporary content. */
public:
	BpiOutStream(std::ostream& os); /**< Constructs an object with specified output stream. */
	virtual ~BpiOutStream(); /**< Default destructor. */
	friend BpiOutStream& operator<<(BpiOutStream &bout, std::string s); /**< Overrides the operator << */
	friend BpiOutStream& operator<<(BpiOutStream &bout, int i); /**< Overrides the operator << */
	friend BpiOutStream& operator<<(BpiOutStream &bout, double d); /**< Overrides the operator << */
	friend BpiOutStream& operator<<(BpiOutStream &bout, uint64_t d); /**< Overrides the operator << */
	friend BpiOutStream& operator<<(BpiOutStream &bout, uint32_t d); /**< Overrides the operator << */
	friend BpiOutStream& operator<<(BpiOutStream &bout, uint16_t d); /**< Overrides the operator << */
	friend BpiOutStream& operator<<(BpiOutStream &bout, uint8_t d); /**< Overrides the operator << */
	friend BpiOutStream& operator<<(BpiOutStream &bout, std::ostream& (*pf)(std::ostream&)); /**< Overrides the operator << */
};

/**
 * \brief Custom debug printer class
 *
 * Implementation of an ostream printer with preceding timestamps. This provides a consistent output for
 * different applications.
 */
class DebugOutStream {
private:
	std::ostream& _os; /**< std::ostream which prints the output. */
	std::stringstream _localStringStream; /**< Storage for local and temporary content. */
public:
	DebugOutStream(std::ostream& os); /**< Constructs an object with specified output stream. */
	virtual ~DebugOutStream(); /**< Default destructor. */
	friend DebugOutStream& operator<<(DebugOutStream &dout, std::string s); /**< Overrides the operator << */
	friend DebugOutStream& operator<<(DebugOutStream &dout, int i); /**< Overrides the operator << */
	friend DebugOutStream& operator<<(DebugOutStream &dout, std::ostream& (*pf)(std::ostream&)); /**< Overrides the operator << */
};

/**
 * \brief Interface between the application and a local stored json file.
 *
 * This class provides trivial mechanisms to read and write key/value pairs from and to json files.
 * Each time a read or write function is called, the application accesses the local file, so that it
 * is possible to modify it at runtime. This class does not provide mechanisms for reading or writing
 * more complex structure than simple key/value pairs.
 *
 */
class Configuration {
public:
	Configuration();
	virtual ~Configuration();
	static std::string getConfig(std::string jsonFilePath, std::string attribute); /**< Determines the value of the given attribute in the json file. */
	static bool setConfig(std::string jsonFilePath, std::string attribute, std::string value); /**< (Over)writes the given attribute/value pair.*/
	static nlohmann::json readJsonFile(std::string pathToFile); /**< Reads a local file into a json object.*/
	static void writeJsonFile(nlohmann::json json, std::string pathToFile); /**< Writes a json object to the filesystem.*/
	static bool deleteConfig(std::string pathToFile, std::string attribute); /**< Deletes a json object.*/
private:
};

/**
 * \brief
 *
 * Simple timer with µs precision.
 */
class Timer {
public:
	Timer() :
			beg_(clock_::now()) {
	}
	void reset() {
		beg_ = clock_::now();
	}
	double elapsed() const {
		return std::chrono::duration_cast<std::chrono::microseconds>(clock_::now() - beg_).count();
	}

private:
	typedef std::chrono::high_resolution_clock clock_;
	std::chrono::time_point<clock_> beg_;
};

/**
 * \brief
 *
 * Simple timer with µs precision.
 */
class Clock {
public:
	Clock();
	void set(uint64_t us);
	void reset();
	double gettime() const;

private:
	typedef std::chrono::high_resolution_clock clock_;
	std::chrono::time_point<clock_> beg_;
};

static std::string bpiName = "------"; /**< Placeholder for the application, which uses this library. */
static std::vector<int> childs; /**< Vector of pid's of forked child processes. */
static Clock localclock;

/*FIXME JN*/
static timespec globalTS;
void write_log(const std::string &text, std::string name);
/*END JN*/

void bpiSetName(std::string name); /**< Overwrites the name placeholder string bpiName. */
std::string bpiGetName();
Clock* getLocalClock();
bool registerChild(int pid); /**< Registers a new forked child by appending its pid to the vector childs. */
bool deregisterChild(int pid); /**< Deregisters a forked child by removing its pid from the vector childs. */
std::vector<int> getAllChilds(); /**< Returns the complete vector of all forked child processes. */
void sigtermHandler(int signum); /**< A handler for the signal SIGTERM. All forked processes in childs will be killed first by sending SIGTERM to them. */
void bpiSystemCall(std::string cmd, int (*printFct)(std::string) = nullptr, int (*printErrFct)(std::string) = nullptr); /**< Performs a sequential system call. Optional pointer to printer functions may be specified as parameter. */
void bpiSystemCall2(std::string cmd, int (*printFct)(std::string) = nullptr, int (*printErrFct)(std::string) = nullptr, std::string cmdId = ""); /**< Performs a concurrent system call.  Optional pointer to printer functions may be specified as parameter.*/
//std::string bpiTimestamp(); /**< Creates a timestamp. */
std::vector<std::string> bpiSplitIntoItems(std::string str, char delimiter); /**< Splits string into a vector of items. */
std::string bpiGetNamestamp(std::string name); /**< Creates a namestamp depending on the specified parameter. */
std::string bpiGetNamestampErr(std::string name); /**< Creates a namestamp depending on the specified parameter. */
std::string bpiGetTimestamp(std::string clocksyncmode); /**< Creates a timestamp. */
uint64_t getLocalTime();
uint64_t getLocalTime48();
std::string bpiConfirm(std::string command, std::string optional); /**< Creates a confirmation for the specified command.*/
bool readInputFromStdin(std::string* s, int bufsize = BUFSIZ);
int readFromFileDescriptor(int fileDescriptor, std::string& data, int buffersize); /**< Reads data from a file descriptor. */
int writeToFileDescriptor(int fileDescriptor, std::string data); /**< Writes data to a file descriptor. */
std::string createCtrlMessage(int argc, ...); /**< Creates a control message. */
std::string createStatCtrlMessage(int argc, ...); /**< Creates a control message. */
std::string macToString(int len, char* mac);
double getCurrentCpuUsage();


#endif /* GLOBALBPILIB_HPP_ */
