/*
 * MessageTypes.cpp
 *
 *      Author: Valentin Fitz
 *   Institute: University of Duisburg-Essen, Germany
 */

#include "MessageTypes.h"

#include "../GlobalBpiLib.hpp"

#include <linux/if_ether.h>
#include <cstring>
#include <string>
#include <sstream>
#include <iostream>
#include <stdlib.h>
#include<stdio.h>
#include <cmath>

Message* createMessageData(const char* data, size_t sizeofdata, size_t mindatasize) {
	Message* msg = new Message;
	if (mindatasize > sizeofdata)
		msg->size = mindatasize;
	else
		msg->size = sizeofdata;
	msg->data = (char*) malloc(msg->size);
	memset(msg->data, 0, msg->size);
	memcpy(msg->data, data, sizeofdata);
	return msg;
}

void deleteMessage(Message* msg) {
	free(msg->data);
	delete (msg);
}

std::string printMessage(Message* msg, bool clear) {
	std::stringstream ss;
	if (clear) {
		char data[msg->size + 1];
		memcpy(data, msg->data, msg->size);
		data[msg->size] = '\0';
		ss << "[ " << data << " ]";
	} else {
		ss << "[ <-- " << msg->size << " --> ]";
	}
	return ss.str();
}

EthernetMessage* createEthernetMessage(char dest[], char source[], short ether_type, Message* data) {
	auto em = new EthernetMessage;
	memcpy(em->destination, dest, ETH_ALEN);
	memcpy(em->source, source, ETH_ALEN);
	em->ether_type = ether_type;
	em->data = data;
	em->fcs = 0;
	em->size = ETH_HLEN + (data == nullptr ? 0 : data->size) + ETH_FCS_LEN;
	em->fcs = generateCrc32(em);
	return em;
}

void deleteMessage(EthernetMessage* msg) {
	deleteMessage(msg->data);
	delete (msg);
}

std::string printMessage(EthernetMessage* msg) {
	std::stringstream ss;
	std::string data = "";
	if (msg->data != nullptr)
		data = printMessage(msg->data);
	ss << "[ " << macToString(6, msg->destination);
	ss << " | " << macToString(6, msg->source);
	ss << " | " << std::setfill('0') << std::setw(2) << std::hex << std::uppercase << msg->ether_type;
	ss << " | " << data;
	ss << " | " << std::setfill('0') << std::setw(2) << std::hex << std::uppercase << msg->fcs << " ]";
	return ss.str();
}

uint32_t generateCrc32(EthernetMessage* msg) {

	uint32_t crc = 0;

	uint8_t vec;
	uint8_t* p;
	uint16_t n;

	for (int i = 0; i < 5; i++) {
		switch (i) {
		case 0:
			p = (uint8_t*) &msg->destination;
			n = 6;
			break;
		case 1:
			p = (uint8_t*) &msg->source;
			n = 6;
			break;
		case 2:
			p = (uint8_t*) &msg->ether_type;
			n = 2;
			break;
		case 3:
			p = (uint8_t*) msg->data->data;
			n = msg->data->size;
			break;
		case 4:
			p = (uint8_t*) &msg->fcs;
			n = (msg->fcs == 0 ? 0 : 4);
			break;
		default:
			break;
		}

		for (int l = n - 1; l >= 0; l--) {
			vec = *((uint16_t*) (p + l));
			for (int i = 7; i >= 0; i--) {
				if (((crc >> 31) & 1) != ((vec >> i) & 1))
					crc = (crc << 1) ^ CRC32GENP;
				else
					crc = (crc << 1);
			}
		}
	}
	return crc;
}

Message* serializeEthernetMessage(EthernetMessage* em) {
	Message* msg = new Message;
	msg->size = em->size;
	msg->data = (char*) malloc(msg->size);
	memset(msg->data, 0, msg->size);
	memcpy(msg->data, em->destination, ETH_ALEN);
	memcpy(msg->data + ETH_ALEN, em->source, ETH_ALEN);
	memcpy(msg->data + ETH_ALEN + ETH_ALEN, &em->ether_type, 2);
	if (em->data != nullptr)
		memcpy(msg->data + ETH_ALEN + ETH_ALEN + 2, em->data->data, em->data->size);
	memcpy(msg->data + em->size - ETH_FCS_LEN, &em->fcs, ETH_FCS_LEN);
	return msg;
}

EthernetMessage* deserializeEthernetMessage(const char* raw, uint16_t sizeofdata) {
	auto em = new EthernetMessage;
	em->size = sizeofdata;
	memcpy(em->destination, raw, ETH_ALEN);
	memcpy(em->source, raw + ETH_ALEN, ETH_ALEN);
	memcpy(&em->ether_type, raw + ETH_ALEN + ETH_ALEN, 2);
	em->data = new Message;
	em->data->size = em->size - ETH_HLEN - ETH_FCS_LEN;
	em->data->data = (char*) malloc(em->data->size);
	memcpy(em->data->data, raw + ETH_HLEN, em->data->size);
	memcpy(&em->fcs, raw + ETH_HLEN + em->data->size, ETH_FCS_LEN);
	return em;
}

EthernetMessage* deserializeEthernetMessage(Message* raw) {
	return deserializeEthernetMessage(raw->data, raw->size);
}

FabanMessage* createFabanMessage(uint8_t senderId, uint8_t sequenceNumber, uint64_t deliveryTime, Message* data) {
	if (deliveryTime >> 48 != 0)
		return nullptr;
	FabanMessage* fabMsg = new FabanMessage;
	fabMsg->senderId = senderId;
	fabMsg->sequenceNumber = sequenceNumber;
	fabMsg->deliveryTime = deliveryTime;
	fabMsg->data = data;
	fabMsg->size = 1 + 1 + 6 + (data == nullptr ? 0 : data->size);
	return fabMsg;
}

void deleteMessage(FabanMessage* msg) {
	deleteMessage(msg->data);
	delete (msg);
}

Message* serializeFabanMessage(FabanMessage* fabMsg) {
	Message* msg = new Message;
	msg->size = fabMsg->size;
	msg->data = (char*) malloc(msg->size);
	memset(msg->data, 0, msg->size);
	memcpy(msg->data, &fabMsg->senderId, 1);
	memcpy(msg->data + 1, &fabMsg->sequenceNumber, 1);
	uint64_t delTime = fabMsg->deliveryTime;
	memcpy(msg->data + 1 + 1, &delTime, 6);
	memcpy(msg->data + 1 + 1 + 6, fabMsg->data->data, fabMsg->data->size);
	return msg;
}

Message* copyMessage(Message* msg) {
	Message* cpy = new Message;
	cpy->size = msg->size;
	cpy->data = (char*) malloc(msg->size);
	memcpy(cpy->data, msg->data, msg->size);
	return cpy;
}

EthernetMessage* copyEthernetMessage(EthernetMessage* em) {
	EthernetMessage* cpy = (EthernetMessage*) malloc(sizeof(EthernetMessage));
	memcpy(cpy, em, sizeof(EthernetMessage));
	cpy->data = copyMessage(em->data);
	return cpy;
}

FabanMessage* deserializeFabanMessage(Message* msg) {
	FabanMessage* fabMsg = new FabanMessage;
	fabMsg->size = msg->size;
	memcpy(&fabMsg->senderId, msg->data, 1);
	memcpy(&fabMsg->sequenceNumber, msg->data + 1, 1);
	uint64_t delTime = 0;
	memcpy(&delTime, msg->data + 1 + 1, 6);
	fabMsg->deliveryTime = delTime;
	fabMsg->data = new Message;
	fabMsg->data->size = fabMsg->size - 1 - 1 - 6;
	fabMsg->data->data = (char*) malloc(fabMsg->data->size);
	memcpy(fabMsg->data->data, msg->data + 1 + 1 + 6, fabMsg->data->size);
	return fabMsg;
}

std::string printMessage(FabanMessage* fabMsg) {
	std::stringstream ss;
	char data[(int) fabMsg->data->size + 1];
	memcpy(data, fabMsg->data, (int) fabMsg->data->size);
	data[fabMsg->data->size] = '\0';
	ss << "[ " << (int) fabMsg->senderId << " | " << (int) fabMsg->sequenceNumber << " | " << (uint64_t) fabMsg->deliveryTime << " | "
			<< printMessage(fabMsg->data) << " ]";
	return ss.str();
}

/**
 * FIXME JN
 */
bool compareEthernetMessage(EthernetMessage* message1, EthernetMessage* message2){
	if(message1 == NULL || message2 == NULL){return false;}

	if( 0 == macToString(6, message1->destination).compare(macToString(6,message2->destination) )  &&
		0 == macToString(6, message1->source).compare(macToString(6,message2->source))&&
		message1->ether_type ==message2->ether_type &&
		0 == strcmp(message1->data->data, message2->data->data) &&
		message1->fcs == message2->fcs
	){
		return true;
	}else{
		return false;
	}
}

