/*
 * MessageTypes.h
 *
 *      Author: Valentin Fitz
 *   Institute: University of Duisburg-Essen, Germany
 */

#ifndef PROTOCOLS_MESSAGETYPES_H_
#define PROTOCOLS_MESSAGETYPES_H_

#include <stdint.h>
#include <string>

#define CRC32GENP 0x04C11DB7
#define FABDATAMIN 38
#define ETHDATAMIN 46

/**
 * \brief Basic message meta information
 */
typedef struct MsgMeta {
} MsgMeta;

/**
 * \brief Ethernet message meta information
 */
typedef struct EthMsgMeta: MsgMeta {
	int port;
} EthMsgMeta;

/**
 * \brief Faban message meta information
 */
typedef struct FabanMeta: EthMsgMeta {
	// TODO
} FabanMeta;

/**
 * \brief Basic message struct from which every other message type is derived from. This struct provides a link to a MsgMeta struct with additional information.
 */
typedef struct ExtMessage {
	uint16_t size; /**< Size in bytes of the complete message (only fields which are transferred over the network). */
	MsgMeta* meta; /**< Pointer to a MsgMeta structure. */
} ExtMessage;

/**
 * \brief Raw message type
 */
typedef struct Message: ExtMessage {
	char* data;
} Message;

/**
 * \brief Ethernet v2 message type
 */
typedef struct EthernetMessage: ExtMessage {
	char destination[6]; /* 6 octets */
	char source[6]; /* 6 octets */
	unsigned short ether_type; /* 2 octets */
	Message* data; /* Data - 46..1500 octets */
	uint32_t fcs; /* 4 octets */
} EthernetMEssage;

/**
 * \brief Faban message type
 */
typedef struct FabanMessage: ExtMessage {
	uint8_t senderId; /* 1 octet */
	uint8_t sequenceNumber; /* 1 octet */
	uint64_t deliveryTime :48; /* 6 octets */
	Message* data; /* Data - 38..1492 octets */
} FabanMessage;

Message* createMessageData(const char* data, size_t sizeofdata, size_t mindatasize = 0);
Message* copyMessage(Message* msg);
void deleteMessage(Message* msg);
std::string printMessage(Message* msg, bool clear=false);

EthernetMessage* createEthernetMessage(char dest[], char source[], short ether_type, Message* data);
bool compareEthernetMessage(EthernetMessage* message1,EthernetMessage* message2);//FIXME JN compare two EthernetMessages
EthernetMessage* copyEthernetMessage(EthernetMessage* em);
void deleteMessage(EthernetMessage* msg);
Message* serializeEthernetMessage(EthernetMessage* em);
EthernetMessage* deserializeEthernetMessage(Message* raw);
EthernetMessage* deserializeEthernetMessage(const char* raw, uint16_t sizeofdata);
std::string printMessage(EthernetMessage* msg);
uint32_t generateCrc32(EthernetMessage* msg);

FabanMessage* createFabanMessage(uint8_t senderId, uint8_t sequenceNumber, uint64_t deliveryTime, Message* data);
void deleteMessage(FabanMessage* msg);
Message* serializeFabanMessage(FabanMessage* fabMsg);
FabanMessage* deserializeFabanMessage(Message* msg);
std::string printMessage(FabanMessage* fabMsg);

#endif /* PROTOCOLS_MESSAGETYPES_H_ */
