/*
 * Port.h
 *
 *      Author: Valentin Fitz
 *   Institute: University of Duisburg-Essen, Germany
 */

#ifndef PROTOCOLS_PORT_H_
#define PROTOCOLS_PORT_H_

#define ACK 0 /*< 1 for ACKs.*/
#define ACKTIMEOUT_US 1000000

#include <pthread.h>
#include <string>
#include <vector>
#include <pcap.h>
#include <linux/if_packet.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "../GlobalBpiLib.hpp"
#include "MessageTypes.h"

class ProtocolLogic;

struct pcapPacket{
	u_char *user;
	const struct pcap_pkthdr* header;
	const u_char *pa;
};

/**
 * \brief Represents an BananaPi R1 ethernet port, starting an ingress port (pcap listener) as well as an egress port (raw socket) thread.
 */
class Port {
public:
	Port(int id, ProtocolLogic* proto);

	BpiOutStream bpiout; /**< Standard output printer. */
	BpiOutStream bpierr; /**< Error output printer. */
	int rxMsg, txMsg; /**< Debugging member variables for retrieving statistics about received and transmitted messages. */

	virtual ~Port();
	void start(); /**< Starts threads for egresss and ingress ports. */
	void stop(); /**< Terminates both running threads. */
	void addMessage(EthernetMessage* msg); /**< Adds message to the outgoing queue. It will be sent as soon as possible. */
	void setNeighbour(char* mac); /**< Sets the destination MAC address (connected device) to the specified parameter. */
	char* getRawMacAddress(); /**< Returns the MAC address of the device as char array. */
	char* getRawDestMacAddress(); /**< Returns the MAC address of the connected device as char array (only correct if it was set before with setNeighbour(...)). */
	std::string getLocalMacAddr(); /**< Returns the MAC address of device as string. */
	int getPortId(); /**< Retrieves the port id as number between 1 and 4. */
	void setLinkSpeed(double d); /**< Sets the linkspeed of the desired transmission speed. */
	void registerEthType(unsigned short eth_type); /**< Registers the specified ethType to this Port. */
//	ProtocolLogic* getProtocolLogic(); /**< Returns a pointer to the PRotocolLogic. */
	uint32_t statTx;
	uint32_t statRx;

private:
	pthread_t _egrPort; /**< Egress port thread. */
	pthread_t _ingPort; /**< Ingress port thread. */
	pthread_t _ingPortFwd;
	pthread_mutex_t _mutex; /**< Mutex for exclusive access on the outgoing message queue. */
	pthread_cond_t _cond; /**< Condition object. */
	pthread_mutex_t _ackMutex;
	pthread_cond_t _ackCond;
	pthread_mutex_t _mutexBuf;
	pthread_cond_t _bufCond;
	int _portId; /**< Id of the port in {1,2,3,4}. */
	std::string _devName; /**< Name of the device in the form 'eth0.x' where x is the _portId. */
	bool active = true; /**< Active flag. */
	char error_buffer[PCAP_ERRBUF_SIZE]; /**< Error buffer. */
	char source_mac[6]; /**< Mac address of this node. */
	char destination_mac[6]; /**< Mac adress of the node connected to this */
	pcap_t *pcap_port = nullptr; /**< Pcap port for message listening (ingress port). */
	std::vector<EthernetMessage*> _msgQueue; /**< Message queue for outgoing messages. */
	std::vector<unsigned short> _regEthTypes; /**< Registered ethTypes. Messages with one of these ethType will be acked. */
	EthernetMessage* _ackMsg;
	int raw_port = -1; /**< File descriptor for the raw port to send on (egress port)*/
	struct sockaddr_ll raw_port_address; /**< Raw port address description. */
	ProtocolLogic* protocol; /**< Pointer to the protocol logic. */
	struct timeval lasttval; /**< Time of the last received message. */
	double transmissionSpeed = 5; /**< Transmission speed in MBit/s. */
	std::vector<struct pcapPacket> _packetBuffer;

	static void * _egrPortThreadEntryFunction(void * This); /** Entry function for the egress port thread. */
	static void * _ingPortThreadEntryFunction(void * This); /** Entry function for the ingress port thread. */
	static void * _ingPortForwarderThreadEntryFunction(void* This);
	void _startEgrPortThread(); /** Logic of the egress port thread. This function does not terminate as long as the port is active. */
	void _startIngPortThread(); /** Logic of the ingress port thread. This function does not terminate as long as the port is active. */
	void _startIngPortForwarderThread();
	static void _packetHandlerCallback(u_char *user, const struct pcap_pkthdr* pkthdr, const u_char* packet); /**< Static callback function for executing the packetHandler function. */
	void _packetHandler(const struct pcap_pkthdr* pkthdr, const u_char* packet); /**< Handler for incoming packets. */
	void _setupPcapListener(); /** Initializes the pcap listener (ingress port). */
	void _setupRawSocket(); /** Initializes the raw socket (egress port). */
	bool _waitForAck(int timeout_us);
	void _sendAck(EthernetMessage* msg);
	void _retrieveMacAddress(); /** Retrieves the mac address of this device. */

	static void* _packetHandlerThreadEntry( void * This );
	void _threadPacketHandler(EthernetMessage* ethmsg);

//	void _putInBuffer(struct pcapPacket);
//	struct pcapPacket _getFromBuffer();

protected:
	virtual bool _handleIncomingPaket(EthernetMessage* msg); /** Virtual function without any function, which can be extended by overriding within a derived class. */
	virtual bool _handleOutgoingPaket(EthernetMessage* msg); /** Virtual function without any function, which can be extended by overriding within a derived class. */
	virtual bool _handleIncomingPaketFI(EthernetMessage* msg); /** Default implementation of a packet handler to offer an easiere possibility to implement fault injection mechanisms in a modular way. */
	virtual bool _handleOutgoingPaketFI(EthernetMessage* msg); /** Default implementation of a packet handler to offer an easiere possibility to implement fault injection mechanisms in a modular way. */
};

#endif /* PROTOCOLS_PORT_H_ */
