/*
 * Port.cpp
 *
 *      Author: Valentin Fitz
 *   Institute: University of Duisburg-Essen, Germany
 */

#include "Port.h"

#include <unistd.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <linux/if_ether.h>
#include <netinet/in.h>
#include <net/if.h>

#include "EtherTypes.h"
#include "ProtocolLogic.h"

using namespace std;

void* Port::_egrPortThreadEntryFunction(void* This) {
	((Port *) This)->_startEgrPortThread();
	return NULL;
}

void Port::_startEgrPortThread() {
	BpiOutStream bpiout(cout);
	BpiOutStream bpierr(cerr);

	stringstream ss;
	ss << "egrPort_" << _portId;

	pthread_setname_np(pthread_self(), ss.str().c_str());

	_setupRawSocket();
	double minTimeNextPacket = 0;
	Timer timer;
	bpiout << "Port egr_" << _portId << " started (" << transmissionSpeed << " MBit/s)" << endl;
	pthread_mutex_lock(&_mutex);
	while (active) {
		pthread_cond_wait(&_cond, &_mutex);
		for (auto message = _msgQueue.begin(); message != _msgQueue.end();) {
			bool send = _handleOutgoingPaketFI(*message);
			if (send) {
				auto msg = serializeEthernetMessage(*message);
				double timeSinceLastPacket = timer.elapsed();
				while (timeSinceLastPacket < minTimeNextPacket) {
					usleep(minTimeNextPacket - timeSinceLastPacket);
					timeSinceLastPacket = timer.elapsed();
				}
				int s = sendto(raw_port, msg->data, msg->size, 0, (struct sockaddr*) &raw_port_address, sizeof(struct sockaddr_ll));
				statTx++;
				timer.reset();
				minTimeNextPacket = (8 * msg->size) / (transmissionSpeed == 0 ? -1 : transmissionSpeed);
				if (s < 0) {
					bpierr << "Sending message failed (egress thread) on port " << getPortId() << endl;
					bpierr << "Failed: " << printMessage(*message) << "\n\t --> " << printMessage((*message)->data, true) << endl;
					minTimeNextPacket = 0;
				} else {
					txMsg++;
				}
				if (ACK) {
					bool ack = _waitForAck(ACKTIMEOUT_US);
					if (!ack)
						bpierr << "Ack missing: " << printMessage(msg) << std::endl;
				}
				deleteMessage(msg);
			}
			free((*message)->data);
			message = _msgQueue.erase(message);
		}
	}
	pthread_mutex_unlock(&_mutex);
	bpiout << "Port egr_" << _portId << " terminated" << endl;

}

void Port::_setupRawSocket() {
	BpiOutStream bpierr(cerr);
	if ((raw_port = socket(AF_PACKET, SOCK_RAW, IPPROTO_RAW)) == -1) {
		bpierr << "Error creating raw socket" << endl;
		active = false;
		return;
	}
	struct ifreq if_idx;
	memset(&if_idx, 0, sizeof(struct ifreq));
	strncpy(if_idx.ifr_name, _devName.c_str(), IFNAMSIZ - 1);
	if (ioctl(raw_port, SIOCGIFINDEX, &if_idx) < 0) {
		bpierr << "Error getting device index (egress thread)" << endl;
		return;
	}
	raw_port_address.sll_ifindex = if_idx.ifr_ifindex;
	raw_port_address.sll_halen = ETH_ALEN;
}

bool Port::_waitForAck(int timeout_us) {
	if (timeout_us == 0)
		return true;

	pthread_mutex_lock(&_ackMutex);
	struct timespec ts;
	timespec_get(&ts, TIME_UTC);

	ts.tv_sec += (int) (timeout_us / 1000000L);
//	bpiout << "Waiting for " << (uint32_t) ts.tv_sec << "s and " << (uint32_t) ts.tv_nsec << "ns" << std::endl;

	if (_ackMsg == nullptr) {
		pthread_cond_timedwait(&_ackCond, &_ackMutex, &ts);
		if (_ackMsg != nullptr) {
			deleteMessage(_ackMsg);
			_ackMsg = nullptr;
			pthread_mutex_unlock(&_ackMutex);
			return true;
		} else {
			pthread_mutex_unlock(&_ackMutex);
			return false;
		}
	} else {
		pthread_mutex_unlock(&_ackMutex);
		return false;
	}
}

bool Port::_handleOutgoingPaket(EthernetMessage* msg) {
	return true;
}

bool Port::_handleOutgoingPaketFI(EthernetMessage*msg) {
	return _handleOutgoingPaket(msg);
}

void Port::addMessage(EthernetMessage* msg) {
	pthread_mutex_lock(&_mutex);
	_msgQueue.push_back(msg);
	pthread_cond_signal(&_cond);
	pthread_mutex_unlock(&_mutex);
}
