/*
 * ProtocolLogic.cpp
 *
 *      Author: Valentin Fitz
 *   Institute: University of Duisburg-Essen, Germany
 */

#include "ProtocolLogic.h"

#include <string>
#include "EtherTypes.h"

ProtocolLogic::ProtocolLogic() :
		protologic(-1), bpiout(std::cout), bpierr(std::cerr) {
	ports = new Port*[4];
	pthread_mutex_init(&_mutex, NULL);
	pthread_cond_init(&_cond, NULL);
}

void ProtocolLogic::start() {
	pthread_create(&protologic, NULL, protoLogicEntry, this);
	pthread_detach(protologic);
}

void ProtocolLogic::_initializeProtocol() {
	std::stringstream ss;
	ss << bpiName << "_logic";
	pthread_setname_np(pthread_self(), ss.str().c_str());

	for (int i = 0; i < 4; i++) {
		ports[i] = _createPort(i + 1);
		ports[i]->setLinkSpeed(8.0);
		ports[i]->start();
	}
	memcpy((void*) source_mac, (void*) (ports[0]->getRawMacAddress()), 6);
}

Port* ProtocolLogic::_createPort(int i) {
	return new Port(i, this);
}

void ProtocolLogic::_executeProtocolLogic() {
	for (auto extMsg = protocol_messages.begin(); extMsg != protocol_messages.end();) {
		EthernetMessage* message = *extMsg;
		std::string msg = std::string(message->data->data);
		bpiout << "Received message: |" << msg << "|" << std::endl;
		bpiout << "NOTE: default behavior of ProtocolLogic::_executeProtocolLogic should be overridden." << std::endl;
		extMsg = protocol_messages.erase(extMsg);
	}
}

void ProtocolLogic::_handleControlMessages() {
	while (!control_messages.empty()) {
		std::string message = control_messages.front();
		control_messages.pop_front();

		std::vector<std::string> items = bpiSplitIntoItems(message, ' ');
		std::string id = items.back();

		items.pop_back();
		std::stringstream ss;
		for (auto item = items.begin(); item != items.end(); item++) {
			ss << *item << " ";
		}

//		bpiout << message << std::endl;
//		if(items[0].compare("_sync") == 0){
//
//		}

		std::string cmd = ss.str().substr(0, ss.str().size() - 1);
		_handleControlMessageFI(id, cmd);
	}
}

void ProtocolLogic::_handleControlMessageFI(std::string cmdId, std::string cmd) {
	_handleControlMessage(cmdId, cmd);
}

void ProtocolLogic::_handleControlMessage(std::string cmdId, std::string cmd) {
	bpiout << "NOTE: default behavior of ProtocolLogic::_handleControlMessage should be overridden." << std::endl;
	bpiout << "NOTE: Sending control message as is from all ports." << std::endl;
	char fake_mac[6] = { 0 };
	for (int i = 0; i < 4; i++) {
		Message* m = createMessageData(fake_mac, 6);
		EthernetMessage* em = createEthernetMessage(fake_mac, ports[i]->getRawMacAddress(), ETH_TYPE_PROTOPROTOCOL, m);
		deleteMessage(m);
		ports[i]->addMessage(em);
	}
}

void ProtocolLogic::_runProtoLogic() {
	_initializeProtocol();
	while (run) {
		pthread_mutex_lock(&_mutex);
		pthread_cond_wait(&_cond, &_mutex);
		_executeProtocolLogic();
		_handleControlMessages();
		pthread_mutex_unlock(&_mutex);
	}
}

void ProtocolLogic::stop() {
	pthread_mutex_lock(&_mutex);
	run = false;
	for (int i = 0; i < 4; i++) {
		ports[i]->stop();
	}
	pthread_cond_signal(&_cond);
	pthread_mutex_unlock(&_mutex);
}

void* ProtocolLogic::protoLogicEntry(void* This) {
	((ProtocolLogic*) This)->_runProtoLogic();
	return nullptr;
}

void ProtocolLogic::portInput(EthernetMessage* input) {
	pthread_mutex_lock(&_mutex);
	protocol_messages.push_back(input);
	pthread_cond_signal(&_cond);
	pthread_mutex_unlock(&_mutex);
}

void ProtocolLogic::controlInput(std::string input) {
	pthread_mutex_lock(&_mutex);
	control_messages.push_back(input);
	pthread_cond_signal(&_cond);
	pthread_mutex_unlock(&_mutex);
}

ProtocolLogic::~ProtocolLogic() {
	stop();
	pthread_mutex_lock(&_mutex);
	for (auto it = protocol_messages.begin(); it != protocol_messages.end(); it++) {
		delete (*it);
		it = protocol_messages.erase(it);
	}
	for (auto it = control_messages.begin(); it != control_messages.end(); it++) {
		it = control_messages.erase(it);
	}
	pthread_mutex_unlock(&_mutex);
}

/**
 * 	DO NOT CALL THIS FUNCTION FOR SENDING MESSAGES!
 *
 *	Adds a message to the egressPorts specified by the binary representation of outPort, i.e.
 *	0<=outPorts = (p3,p2,p1,p0) <=15 where pi=1 indicates to send on port i.
 *
 *	Example: outPorts=0b0101 means to send on port 0 as well as port 2.
 */
void ProtocolLogic::_sendMessage(EthernetMessage* msg, int outPorts) {
	if (outPorts < 0 || outPorts > 0b1111) {
		std::stringstream ss;
		ss << "_sendMessage: Invalid parameter (outPorts) specified [" << outPorts << "]";
		bpierr << ss.str() << std::endl;
		return;
	}
	int p;
	for (int i = 0; i < 4; i++) {
		p = outPorts % 2;
		outPorts = outPorts / 2;
		if (p == 1) {
			ports[i]->addMessage(copyEthernetMessage(msg));
		}
	}
	deleteMessage(msg);
}

/**
 *	Adds a message to the egressPorts specified by the binary representation of outPort, i.e.
 *	0<=outPorts = (p3,p2,p1,p0) <=15 where pi=1 indicates to send on port i.
 *
 *	Example: outPorts=0b0101 means to send on port 0 as well as port 2.
 */
void ProtocolLogic::sendMessage(EthernetMessage* msg, int outPorts) {
	_sendMessage(msg, outPorts);
}

