#include "LsInterface.h"

#include <iostream>
#include <unistd.h>
#include <chrono>
#include <signal.h>
#include <sys/wait.h>
#include <sstream>
#include <cstring>
#include <sys/select.h>
#include <fcntl.h>

using namespace std;

LsInterface::LsInterface(string uid, string runPath) :
		_uid(uid), _runPath(runPath), _pid(-1), _active(false), _piThread(-1), _fdChildIn(-1), _fdChildOut(-1), _fdChildErr(-1) {
}

LsInterface::~LsInterface() {
	if (_pid != -1)
		kill(_pid, SIGKILL);
	close(_fdChildIn);
	close(_fdChildOut);
	close(_fdChildErr);
}

void LsInterface::start() {
	if (!_active) {
		_active = true;
		pthread_create(&_piThread, NULL, _threadEntryFunction, this);
		pthread_detach(_piThread);
	} else {
		printToTcpConnection("Local Service Already running");
	}
}

void LsInterface::stopLocalService() {
	if (_active) {
		_active = false;
		if (_pid) {
			kill(_pid, SIGTERM);
			if (deregisterChild(_pid))
				dout << "Successfully deregistered child with pid " << _pid << endl;
			else
				derr << "Deregistration failed. There is no registered process with pid " << _pid << endl;
		}
		pthread_cancel(_piThread);
		printToTcpConnection("LocalService stopped");
		signal(SIGCHLD, SIG_IGN);
	} else {
		printToTcpConnection("Nothing to stop. LocalService is not running!");
	}
}

int LsInterface::sendToLocalService(std::string s) {
	s.append("\n");
	return write(_fdChildIn, s.c_str(), s.length());
}

bool LsInterface::isActive() {
	return _active;
}

std::string LsInterface::getUid() {
	return _uid;
}

int* LsInterface::getFdLsOutput() {
	return &_fdChildOut;
}

int* LsInterface::getFdLsOutputErr() {
	return &_fdChildErr;
}

void* LsInterface::_threadEntryFunction(void* This) {
	((LsInterface *) This)->_startThread();
	return NULL;
}

void LsInterface::_startThread() {
	std::stringstream ss;
	ss << "Interface for \"" << _uid << "\" started." << endl;
	printToTcpConnection(ss.str());

	pthread_setname_np(pthread_self(), "LsInterface");

	if (_executeProtocol() < 0) {
		derr << "Execution of the LocalService failed!" << endl;
		printErrToTcpConnection("Execution of the LocalService failed!");
		this->stopLocalService();
	}

	int status;

	while (_active) {
		if (waitpid(_pid, &status, WUNTRACED) < 0) {
			derr << "LocalService terminated or killed." << endl;
			printErrToTcpConnection(ss.str());
			this->stopLocalService();
		}
	}
}

int LsInterface::_executeProtocol() {
	int fd_in[2];
	int fd_out[2];
	int fd_err[2];

	if (pipe(fd_in) < 0 || pipe(fd_out) < 0 || pipe(fd_err) < 0) {
		derr << "Error creating pipe. Protocol \"" << _uid << "\" could not be started." << endl;
		return -1;
	}

	if ((_pid = fork()) < 0) {
		derr << "Error forking process. Protocol \"" << _uid << "\" could not be started." << endl;
		return -1;
	} else if (_pid == 0) { // Child Process
		close(fd_in[1]);
		close(fd_out[0]);
		close(fd_err[0]);

		if (dup2(fd_in[0], fileno(stdin)) < 0 || dup2(fd_out[1], fileno(stdout)) < 0 || dup2(fd_err[1], fileno(stderr)) < 0) {
			printErrToTcpConnection("Error duplicating files descriptors. LocalService could not be started.");
			return -1;
		}

		close(fd_in[0]);
		close(fd_out[1]);
		close(fd_err[1]);

		std::stringstream ss;
		ss << (long long int) getLocalClock()->gettime();

		if (execl(_runPath.c_str(), _uid.c_str(), ss.str().c_str(), NULL) == -1) {
			BpiOutStream bpierrChild(cerr);
			printErrToTcpConnection("Error executing LocalService.");
			_Exit(-1);
		}
	} else { //Parent Process
		if (registerChild(_pid))
			dout << "Successfully registered child process with pid " << _pid << endl;
		else
			derr << "Registration of child process with pid " << _pid << " failed. PID already registered" << endl;
		close(fd_in[0]);
		close(fd_out[1]);
		close(fd_err[1]);

		_fdChildIn = fd_in[1];
		_fdChildOut = fd_out[0];
		_fdChildErr = fd_err[0];
	}

	return 0;
}
