#ifndef TCPINTERFACE_H_
#define TCPINTERFACE_H_

#define COMMAND_GETCONFIG "getconfig"
#define COMMAND_SETCONFIG "setconfig"
#define COMMAND_STATUS "status"
#define COMMAND_COMPILE "compile"
#define COMMAND_COMPILELIB "compilelib"
#define COMMAND_CLEAN "clean"
#define COMMAND_CLEANLIB "cleanlib"
#define COMMAND_START "start"
#define COMMAND_STOP "stop"
#define COMMAND_ALIVE "<alive>"
#define COMMAND_ADIEU "adieu"
#define COMMAND_SHUTDOWN "shutdown"
#define COMMAND_EXIT "exit"
#define COMMAND_SSH "ssh"

//FIXME JN
#define COMMAND_SYNC "_sync"
#define TEN_POWER_NINE 1000000000
//#define COMMAND_SYNC "_syncsystem"
//END JN

#define ACKMODE_NONE 0
#define ACKMODE_ACKWRAP 1
#define ACKMODE_ACKHASHWRAP 2

#include "LocalBpiLib.h"
#include "LsInterface.h"
#include <string>
#include <map>
#include <vector>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <pthread.h>

/**
 * \brief Connection between a tcp client and the LocalService.
 *
 * This class represents an interface between a tcp client and the concurrent thread LsInterface, which communicates
 * with an instance of the LocalService. It listens to input from the tcp client and the LocalService, respectivly, and
 * implements a command handling and a forwarding mechanism.
 */
class TcpInterface {
public:
	int _fdTcpBd = -1; /**< FileDescriptor. */
	bool active = true; /**< Indicates whether the IoService is active or not. Setting this to false will cause a clean termination of the application. */
	bool connected = false; /**< Indicates whether a tcp connection exists or not. */
	pthread_cond_t _cond; /**< Waiting condition, which is used by the UDP connector to go into a standby phase. */
	pthread_mutex_t _mutex; /**< Mutex for access at common variables. Used in combination with wait conditions. */

	TcpInterface(std::string execPath);
	virtual ~TcpInterface();

	void start(); /**< Obtains and handles in an infinite loop incoming commands. */
	bool setLsInterface(LsInterface* __lsIntfce); /**< Sets a pointer to a specified LsInterface object. */
	void repeatFdListening(); /**< Interrupts the select statement which causes an iteration of the fd listening. */

private:
	LsInterface* _lsIntfce; /**< Pointer to the LsInterface. */
	pthread_t _piThread; /**< Thread object. */
	pthread_t _csThread; /**< Thread object for clock synchronization. */

	int _ackmode; /**< Indicates the mode of acknowledgements. */
	int _fdCtrlIn; /**< Control pipe in. */
	int _fdCtrlOut; /**< Control pipe out. */
	int _buffersize = 500; /**< Buffersize. */
	std::string _clocksyncmode;
	long long int _old = -1;
	long long int _clockSyncTxTimepoint[CLOCKSYNCVALUECOUNT][2];

	/*FIXME JN
	 * Variables are created here
	 * because this saves time when running the clock synchronisation
	 * since no new variables have to be initialized
	 */
	timespec localtimeTP;
	timespec _oldTP;
	timespec offsetTP;
	timespec diffTP;
	timespec corrTP;
	long long int localtimeNano =-1;
	long long int _oldNano =-1;
	long long int offsetNano,diffNano;
	/*END JN */

	bool csInitFinished = false;
	uint8_t _csPointer = 0;
	int* _fdTcpOut; /**< Pointer to a filedescriptor, from which one may read data coming from the LsInterface. */
	int* _fdTcpErr; /**< Pointer to a filedescriptor, from which one may read error messages coming from the LsInterface. */
	std::string _partialMsgBuffer = ""; /**< Temporary storage for message concatenation. */
	std::string _partialMsgChildBuffer = ""; /**< Temporary storage for message concatenation. */
	std::string _partialMsgChildBufferErr = ""; /**< Temporary storage for message concatenation. */
	std::string _execPath;

	static void * _threadEntryFunction(void * This); /**< Entry function for thread execution. */
	static void * _clocksyncThreadEntryFunction(void * This); /**< Entry function for the clock sync thread. */
	void _startThread(); /**< Core of the thread. */
	void _clockSync(); /**< Clock synchronisation function, which is executed within a separate thread while the TcpInterface Thread is active. */
	void _handleInput(std::string input); /**< Realizes command handling. */
	void _PrintIncomingMessage(std::string s); /**< Custom print function for messages. */
	void _PrintOutgoingMessage(std::string s); /**< Custom print function for messages. */
	int _sendTcpMessage(std::string s); /**< Sends the specified string over the tcp connection. */
	int _sendTcpMessageWithNamestamp(std::string s); /**< Sends the specified string with a namestamp over the tcp connection. */

	void _CommandGetConfig(std::string cmdId, std::string input); /**< Implementation of the command COMMAND_GETCONFIG. */
	void _CommandSetConfig(std::string cmdId, std::string input); /**< Implementation of the command COMMAND_SETCONFIG. */
	void _CommandStatus(std::string cmdId, std::string input); /**< Implementation of the command COMMAND_STATUS. */
	void _CommandCompile(std::string cmdId, std::string input); /**< Implementation of the command COMMAND_COMPILE. */
	void _CommandCompilelib(std::string cmdId, std::string input); /**< Implementation of the command COMMAND_COMPILELIB. */
	void _CommandClean(std::string cmdId, std::string input); /**< Implementation of the command COMMAND_CLEAN. */
	void _CommandCleanlib(std::string cmdId, std::string input); /**< Implementation of the command COMMAND_CLEANLIB. */
	void _CommandStart(std::string cmdId, std::string input); /**< Implementation of the command COMMAND_START. */
	void _CommandStop(std::string cmdId, std::string input); /**< Implementation of the command COMMAND_STOP. */
	void _CommandAdieu(std::string input); /**< Implementation of the command COMMAND_ADIEU. */
//	void _CommandSync(std::string cmdId, std::string input); /**< Implementation of the command COMMAND_SYNC. */
	void _CommandShutdown(std::string id, std::string input); /**< Implementation of the command COMMAND_SHUTDOWN. */
	void _CommandExit(); /**< Implementation of the command COMMAND_EXIT. */
	void _CommandSsh(std::string cmdId, std::string input);

};

#endif /* TCPINTERFACE_H_ */
