/*
 * CommandHandler.cpp
 *
 *  Created on: Jul 28, 2017
 *      Author: valentin
 */

#include "InputHandler.h"
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <iostream>
#include <string.h>
#include <stdio.h>
#include <algorithm>
#include <signal.h>
#include <limits.h>

using namespace std;
using json=nlohmann::json;

InputHandler::InputHandler(std::string execPath) :
		bpiout(cout), bpierr(cerr), _execPath(execPath) {

}

/**
 * Reads Comands from STDIN and initiates a command handling.
 */
void InputHandler::runLoop() {
	char cmd[PIPE_BUF];

	while (_run)
		if (fgets(cmd, PIPE_BUF, stdin) != NULL)
			handleCommand(cmd);
}

/**
 * Detects which command has to be started and initiate corresponding function calls. All known commands are
 * defined in InputHandler.h. For a command "exampleCommand" there should exist a
 * <b>\#define COMMAND_EXAMPLECOMMAND exampleCommand</b> which helps to identify known commands. After recognizing
 * a command, a function call with the naming convention <b>_CommandExampleCommand(input)</b> should be the only
 * operation to be done in this function. If a command could not be recognized, the function checks whether this
 * could be a command for an active Protocol. If this is true, the function sends the command to the corresponding
 * Protocol. Otherwise the function results in an error output.
 * @param cmdIn Command to be handled as char sequence
 */
void InputHandler::handleCommand(char *cmdIn) {

	cmdIn[strlen(cmdIn) - 1] = 0;
	string input(cmdIn);
	std::vector<std::string> items = bpiSplitIntoItems(input, ' ');
	string cmdId = "";
	if (items.size() < 2 || strncmp("ID", items.at(items.size() - 1).c_str(), 2) != 0) {
		bpiout << createStatCtrlMessage(3, cmdId.c_str(), "error", "command got no id") << endl;
		return;
	}
	cmdId = items.back();
	items.pop_back();

	std::stringstream ss;
	for (auto item = items.begin(); item != items.end(); item++) {
		ss << *item << " ";
	}
	std::string cmd = ss.str().substr(0, ss.str().size() - 1);
	if (strncmp(COMMAND_EXIT, cmdIn, strlen(COMMAND_EXIT)) == 0) {
		_CommandExit(cmdId);
	} else if (strncmp(COMMAND_SYNC, cmdIn, strlen(COMMAND_SYNC)) == 0) {
		_CommandSyncClock(cmdId, cmd);
	} else if (strncmp(COMMAND_GETCONFIG, cmdIn, strlen(COMMAND_GETCONFIG)) == 0) {
		_CommandGetConfig(cmdId, cmd);
	} else if (strncmp(COMMAND_SETCONFIG, cmdIn, strlen(COMMAND_SETCONFIG)) == 0) {
		_CommandSetConfig(cmdId, cmd);
	} else if (strncmp(COMMAND_DELETECONFIG, cmdIn, strlen(COMMAND_DELETECONFIG)) == 0) {
		_CommandDeleteConfig(cmdId, cmd);
	} else if (strncmp(COMMAND_COMPILE, cmdIn, strlen(COMMAND_COMPILE)) == 0) {
		_CommandCompile(cmdId, cmd);
	} else if (strncmp(COMMAND_START, cmdIn, strlen(COMMAND_START)) == 0) {
		_CommandStart(cmdId, cmd);
	} else if (strncmp(COMMAND_STOP, cmdIn, strlen(COMMAND_STOP)) == 0) {
		_CommandStop(cmdId, cmd);
	} else if (strncmp(COMMAND_RESTART, cmdIn, strlen(COMMAND_RESTART)) == 0) {
		_CommandRestart(cmdId, cmd);
	} else if (strncmp(COMAMND_STATUS, cmdIn, strlen(COMAMND_STATUS)) == 0) {
		_CommandStatus(cmdId, cmd);
	} else if (strncmp(COMMAND_SETSYSTEMCLOCK, cmdIn, strlen(COMMAND_SETSYSTEMCLOCK)) == 0) {
		_CommandSetSystemClock(cmdId, cmd);
}	else {
		if (_activeProtocols.count(items[0]) > 0) {
			ProtocolInterface* pi = _activeProtocols.find(items[0])->second;
			if (pi->isActive()) {
				bpiout << createStatCtrlMessage(2, cmdId.c_str(), "success") << std::endl;
				pi->send(string(cmdIn).substr(items[0].size() + 1, std::string::npos));
			} else {
				bpierr << "Protocol \"" << items[0] << "\" found but is not active." << endl;
			}
			return;
		}
		bpierr << "Unknown command " << "\"" << items[0] << "\" received." << endl;
		bpiout << createStatCtrlMessage(4, cmdId.c_str(), "error", "not a command or protocol name", cmdIn) << std::endl;
		return;
	}
}

void InputHandler::_CommandGetConfig(std::string id, std::string input) {
	std::vector<std::string> items = bpiSplitIntoItems(input, ' ');
	if (items.size() == 1) {
		json j = Configuration::readJsonFile(JSON_CONFIGFILE);
		if (j.empty()) {
			bpiout << "Config file is empty" << endl;
		} else {
			stringstream tmpSs;
			for (json::iterator it = j.begin(); it != j.end(); ++it) {
				tmpSs.str("");
				tmpSs << "\"" << it.key() << "\" : " << it.value() << endl;
				bpiout << tmpSs.str().c_str() << endl;
			}
		}
		bpiout << createStatCtrlMessage(2, id.c_str(), "success") << endl;
	} else {
		std::stringstream ss;
		string value = Configuration::getConfig(JSON_CONFIGFILE, items.at(1));
		if (value.size() == 0) {
			bpiout << "Parameter is not existing" << endl;
			bpiout << createStatCtrlMessage(2, id.c_str(), "error") << std::endl;
		} else {
			ss << "\"" << items.at(1) << "\": \"" << value << "\"" << endl;
			bpiout << ss.str() << endl;
			bpiout << createStatCtrlMessage(2, id.c_str(), "success") << std::endl;
		}
	}
}

void InputHandler::_CommandSetConfig(std::string id, std::string input) {
	std::vector<std::string> items = bpiSplitIntoItems(input, ' ');
	if (items.size() < 3) {
		bpierr << (std::string("Invalid key/value assignment: \"").append(input).append("\"")) << endl;
		bpiout << createStatCtrlMessage(3, id.c_str(), "error", "Invalid key/value assignment");
	} else {
		bool result = Configuration::setConfig(JSON_CONFIGFILE, items.at(1), items.at(2));
		if (result) {
			std::stringstream ss;
			ss << "Successfully set configuration: \"" << items.at(1) << "\": \"" << items.at(2) << "\"" << endl;
			bpiout << (ss.str()) << endl;
			bpiout << createStatCtrlMessage(2, id.c_str(), "success") << endl;
		} else {
			bpierr << (std::string("Configuration could not be set: \"").append(input).append("\"")) << endl;
			bpiout << createStatCtrlMessage(3, id.c_str(), "error", "Configuration could not be set");
		}
	}
}

void InputHandler::_CommandDeleteConfig(std::string id, std::string input) {
	std::vector<std::string> items = bpiSplitIntoItems(input, ' ');
	if (items.size() == 1) {
		bpierr << "Config key was not specified" << endl;
		bpiout << createStatCtrlMessage(2, id.c_str(), "error") << endl;
	} else {
		int num = 0;
		for (size_t i = 1; i < items.size(); i++) {
			if (Configuration::deleteConfig(JSON_CONFIGFILE, items.at(i)))
				num++;
		}
		bpiout << "Deleted " << num << " attributes" << endl;
		bpiout << createStatCtrlMessage(2, id.c_str(), "success") << std::endl;
	}
}

void InputHandler::_CommandCompile(std::string id, std::string input) {
	vector<string> items = bpiSplitIntoItems(input, ' ');
	if (items.size() <= 1) {
		bpierr << "Compilation failed. No protocol name specified." << endl;
		bpierr << "Execution failed. No protocol name specified." << endl;
		return;
	} else if (items.size() > 3) {
		bpierr << "Compilation failed. Too many Arguments (\"" << input << "\")." << endl;
		bpiout << createStatCtrlMessage(3, id.c_str(), "error", "too many arguments") << std::endl;
		return;
	} else {
		string path = Configuration::getConfig(JSON_CONFIGFILE, items.at(1));
		if (path.empty()) {
			bpierr << "Compilation failed. Could not find a path in the configuration file for protocol \"" << items.at(1) << "\"" << endl;
			bpiout << createStatCtrlMessage(3, id.c_str(), "error", "could not find protocol") << std::endl;
			return;
		}
		path.append("/");
		if (chdir(path.c_str()) != 0) {
			bpierr << "Compilation failed. \"" << items.at(1) << "\" is not a directory." << endl;
			bpiout << createCtrlMessage(3, id.c_str(), "error", "could not change into protocol directory") << std::endl;
			return;
		}
		bpiSystemCall("make all");
		bpiout << createStatCtrlMessage(2, id.c_str(), "success") << std::endl;
	}
}

void InputHandler::_CommandStart(std::string id, std::string input) {
	vector<string> items = bpiSplitIntoItems(input, ' ');
	if (items.size() <= 1) {
		bpierr << "Execution failed. No protocol name specified." << endl;
		bpiout << createStatCtrlMessage(3, id.c_str(), "error", "no protocol name specified") << std::endl;
		return;
	} else if (items.size() > 3) {
		bpierr << "Execution failed. Too many Arguments (\"" << input << "\")." << endl;
		bpiout << createStatCtrlMessage(3, id.c_str(), "error", "too many arguments") << std::endl;
		return;
	} else {
		string protocolName = items.at(1);
		string protocolExecutionPath = Configuration::getConfig(JSON_CONFIGFILE, protocolName);
		if (protocolExecutionPath.empty()) {
			bpierr << "Execution failed. Could not find a path in the configuration file for protocol \"" << protocolName << "\"" << endl;
			bpiout << createStatCtrlMessage(3, id.c_str(), "error", "could not find protocol") << std::endl;
			return;
		}
		protocolExecutionPath.append("/");
		protocolExecutionPath.append(protocolName);

		if (_activeProtocols.count(protocolName) > 0) {
			ProtocolInterface* pi = _activeProtocols.find(protocolName)->second;
			if (pi == NULL || !(pi->isActive())) {
				delete (pi);
				_activeProtocols.erase(protocolName);
				bpiout << createStatCtrlMessage(2, id.c_str(), "success") << std::endl;
			} else if (pi != NULL && pi->isActive()) {
				bpierr << "Protocol \"" << protocolName << "\" already running. Terminate before start." << endl;
				bpiout << createStatCtrlMessage(3, id.c_str(), "error", "protocol already running") << std::endl;
				return;
			}
		}

		ProtocolInterface* pi = new ProtocolInterface(protocolName, protocolExecutionPath);
		pi->start();
		_activeProtocols.insert(pair<string, ProtocolInterface*>(protocolName, pi));
		bpiout << createStatCtrlMessage(2, id.c_str(), "success") << std::endl;
	}
}

void InputHandler::_CommandStop(std::string id, std::string input) {
	vector<string> items = bpiSplitIntoItems(input, ' ');
	if (items.size() <= 1) {
		bpierr << "Execution failed. No protocol name specified." << endl;
		bpiout << createStatCtrlMessage(3, id.c_str(), "error", "no protocol name specified") << std::endl;
		return;
	} else if (items.size() > 3) {
		bpierr << "Execution failed. Too many Arguments (\"" << input << "\")." << endl;
		bpiout << createStatCtrlMessage(3, id.c_str(), "error", "too many arguments") << std::endl;
		return;
	} else {
		string protocolName = items.at(1);
		if (_activeProtocols.count(protocolName) < 1) {
			bpierr << "Nothing to terminate. Protocol \"" << protocolName << "\" not running." << endl;
			bpiout << createStatCtrlMessage(3, id.c_str(), "error", "protocol not running") << std::endl;
		} else {
			ProtocolInterface* p = _activeProtocols.find(protocolName)->second;
			_activeProtocols.erase(protocolName);
			p->stop();
			delete (p);
			bpiout << createStatCtrlMessage(2, id.c_str(), "success") << std::endl;
		}
	}
}

void InputHandler::_CommandRestart(std::string id, std::string input) {
	_CommandStop("ID000", input); // TODO
	_CommandCompile("ID000", input);
	_CommandStart("ID000", input);
	bpiout << createStatCtrlMessage(2, id.c_str(), "success") << std::endl;
}

void InputHandler::_CommandStatus(std::string id, std::string input) {
	struct stat t_stat;
	stat(_execPath.c_str(), &t_stat);
	struct tm * timeinfo = localtime(&t_stat.st_ctime);

	stringstream ss;
	string lm = "     ";
	string end = "";
	if (_activeProtocols.size() > 0) {
		lm = "   \u2551  ";
		end = "   \u2551\n";
	}
	ss << "\u255A\u2550 Local Service [ACTIVE]\n" << lm << asctime(timeinfo) << end;
	size_t i = 0;
	size_t totalnumber = _activeProtocols.size();
	for (std::map<string, ProtocolInterface*>::iterator it = _activeProtocols.begin(); it != _activeProtocols.end(); ++it) {
		string pre;
		string mid;
		if (++i == totalnumber) {
			pre = "   \u255A\u2550 ";
			mid = "        ";
		} else {
			pre = "   \u2560\u2550 ";
			mid = "   \u2551    ";
		}
		if (it->second->isActive()) {
			stat(it->second->getRunPath().c_str(), &t_stat);
			timeinfo = localtime(&t_stat.st_ctime);
			ss << pre << it->first << " [ACTIVE]\n" << mid << asctime(timeinfo) << mid << endl;
		} else {
			ss << pre << it->first << " [INACTIVE]\n" << mid << "ProtocolInterface deleted!\n" << mid << endl;
			_activeProtocols.erase(it->first);
			delete (it->second);
		}
	}
	bpiout << ss.str() << endl;
	bpiout << createStatCtrlMessage(2, id.c_str(), "success") << std::endl;
}

void InputHandler::_CommandSyncClock(std::string id, std::string input) {

	std::vector<std::string> items = bpiSplitIntoItems(input, ' ');
	long long int time = std::stoull(items[1]);
	getLocalClock()->set(time);
//	bpiout << "Synced" << std::endl;
	for (std::map<string, ProtocolInterface*>::iterator it = _activeProtocols.begin(); it != _activeProtocols.end(); ++it) {
		if (it->second->isActive()) {
			std::stringstream ss;
			ss << "_sync " << (long long int) getLocalClock()->gettime() << " IDxxxx";
			it->second->send(ss.str());
//			bpiout << "Synced " << it->second->getUid() << std::endl;
		}
	}
}

void InputHandler::_CommandExit(std::string id) {
	bpiout << createStatCtrlMessage(2, id.c_str(), "success") << std::endl;
	bpiout << "Local Service terminating" << endl;
	_run = false;
}


void InputHandler::_CommandSetSystemClock(std::string id, std::string input){
	std::vector<std::string> items = bpiSplitIntoItems(input, ' ');
		long long int time = std::stoull(items[1]);
		timespec set;
		set.tv_sec  = time/1000000000;
		set.tv_nsec = time%1000000000;
		clock_settime(CLOCK_REALTIME, &set);
		bpiout << createStatCtrlMessage(2, id.c_str(), "success") << std::endl;

}
