/*
 * ProtocolInterface.cpp
 *
 *  Created on: Aug 31, 2017
 *      Author: valentin
 */

#include "ProtocolInterface.h"
#include <iostream>
#include <unistd.h>
#include <chrono>
#include <signal.h>
#include <sys/wait.h>
#include <sstream>
#include <cstring>
#include <sys/select.h>
#include <fcntl.h>

using namespace std;

ProtocolInterface::ProtocolInterface(string uid, string runPath) :
		bpiout(cout), bpierr(cerr), _uid(uid), _runPath(runPath), _pid(-1), _active(true), _piThread(-1), _fdIntfceIn(-1), _fdIntfceOut(-1), _fdChildIn(-1), _fdChildOut(-1), _fdChildErr(
				-1) {
}

ProtocolInterface::~ProtocolInterface() {
	if (_pid != -1)
		kill(_pid, SIGKILL);
	close(_fdIntfceIn);
	close(_fdIntfceOut);
	close(_fdChildIn);
	close(_fdChildOut);
	close(_fdChildErr);
}

void ProtocolInterface::start() {
	pthread_create(&_piThread, NULL, _threadEntryFunction, this);
	pthread_detach(_piThread);
}

void ProtocolInterface::stop() {
	_active = false;
	bpiout << "Interface for \"" << _uid << "\" stopped." << endl;
	pthread_cancel(_piThread);
	signal(SIGCHLD, SIG_IGN);
}

int ProtocolInterface::send(std::string s) {
	s.append("\n");
	return write(_fdIntfceIn, s.c_str(), s.length());
}

bool ProtocolInterface::isActive() {
	return _active;
}

std::string ProtocolInterface::getUid() {
	return _uid;
}

void* ProtocolInterface::_threadEntryFunction(void* This) {
	((ProtocolInterface *) This)->_startThread();
	return NULL;
}

void ProtocolInterface::_startThread() {

	bpiout << "Interface for \"" << _uid << "\" started." << endl;
	pthread_setname_np(pthread_self(), (_uid + "_intfce").c_str());

	if (_executeProtocol() < 0) {
		bpierr << _pid << ": Protocol execution for \"" << _uid << "\" failed! Kill child process and stop Interface." << endl;
		this->stop();
	}

	fd_set readfds;
	string strBuffer;
	int status;

	while (_active) {
		if (waitpid(_pid, &status, WNOHANG) < 0) {
			bpierr << "Protocol \"" << _uid << "\" terminated or killed." << endl;
			this->stop();
		}

		FD_ZERO(&readfds);
		FD_SET(_fdChildOut, &readfds);
		FD_SET(_fdChildErr, &readfds);
		FD_SET(_fdIntfceOut, &readfds);

		if (select(max(max(_fdChildOut, _fdChildErr), _fdIntfceOut) + 1, &readfds, NULL, NULL, NULL) < 0) {
			bpierr << "Failed reading file descriptors of Protocol \"" << _uid << "\"." << endl;
		}

		if (FD_ISSET(_fdChildOut, &readfds))
			if (_read(_fdChildOut, strBuffer) > 0)
				_handleChildOutput(strBuffer);

		if (FD_ISSET(_fdChildErr, &readfds))
			if (_read(_fdChildErr, strBuffer) > 0)
				_handleChildError(strBuffer);

		if (FD_ISSET(_fdIntfceOut, &readfds))
			if (_read(_fdIntfceOut, strBuffer) > 0) {
				_write(_fdChildIn, strBuffer);
			}
	}
}

int ProtocolInterface::_executeProtocol() {
	int fd_in[2];
	int fd_out[2];
	int fd_err[2];

	if (pipe(fd_in) < 0 || pipe(fd_out) < 0 || pipe(fd_err) < 0) {
		bpierr << "Error creating pipe. Protocol \"" << _uid << "\" could not be started." << endl;
		return -1;
	}

	if ((_pid = fork()) < 0) {
		bpierr << "Error forking process. Protocol \"" << _uid << "\" could not be started." << endl;
		return -1;
	} else if (_pid == 0) { // Child Process
		close(fd_in[1]);
		close(fd_out[0]);
		close(fd_err[0]);

		if (dup2(fd_in[0], fileno(stdin)) < 0 || dup2(fd_out[1], fileno(stdout)) < 0 || dup2(fd_err[1], fileno(stderr)) < 0) {
			bpierr << "Error duplicating files descriptors. Protocol \"" << _uid << "\" could not be started." << endl;
			return -1;
		}

		close(fd_in[0]);
		close(fd_out[1]);
		close(fd_err[1]);

		std::stringstream ss;
		ss << (long long int) getLocalClock()->gettime();

		if (execl(_runPath.c_str(), _uid.c_str(), ss.str().c_str(), NULL) == -1) {
			BpiOutStream bpierrChild(cerr);
			bpierrChild << "Error executing protocol \"" << _uid << "\"." << endl;
			_Exit(-1);
		}
	} else { //Parent Process

		if (registerChild(_pid))
			bpiout << "Successfully registered child process with pid " << _pid << endl;
		else
			bpierr << "Registration of child process with pid " << _pid << " failed. PID already registered" << endl;

		close(fd_in[0]);
		close(fd_out[1]);
		close(fd_err[1]);

		int fd_local[2];
		if (pipe(fd_local) < 0) {
			bpierr << "Could not create pipe for interface IO." << endl;
			return -1;
		}
		_fdIntfceIn = fd_local[1];
		_fdIntfceOut = fd_local[0];

		_fdChildIn = fd_in[1];
		_fdChildOut = fd_out[0];
		_fdChildErr = fd_err[0];
	}

	return 0;
}

int ProtocolInterface::_write(int fileDescriptor, std::string data) {
	std::string s = data;
	if (s.at(s.length() - 1) != '\n')
		s.append("\n");
	return write(fileDescriptor, s.c_str(), s.length());

}

int ProtocolInterface::_read(int fileDescriptor, std::string& data) {
	char b[BUFFERSIZE];
	stringstream ss;
	memset(&b[0], 0, sizeof(b));
	int n = -1;
	if ((n = read(fileDescriptor, b, sizeof(b) - 1)) > 0) {
		b[n] = '\0';
		ss << b;
		*&data = ss.str();
	}
	return n;
}

/**
 * This function just prints child output to STDOUT by default. It is intended for the extension of more complex behaviour
 * for received output, e.g. filter granularity for statistics and monitoring messages.
 * @param out Output string from child.
 */
void ProtocolInterface::_handleChildOutput(std::string out) {
	cout << "" << out << "" << flush;
}

/**
 * This function just prints child errors to STDERR by default. It is intended for the extension of more complex behaviour
 * for received output, e.g. filter granularity for statistics and monitoring messages.
 * @param err Error string from child.
 */
void ProtocolInterface::_handleChildError(std::string err) {
	cerr << err << flush;
}

std::string ProtocolInterface::getRunPath() {
	return _runPath;
}
