/*
 * ProtocolInterface.h
 *
 *  Created on: Aug 31, 2017
 *      Author: valentin
 */

#ifndef PROTOCOLINTERFACE_H_
#define PROTOCOLINTERFACE_H_

#define BUFFERSIZE GEN_BUFFERSIZE

#include "LocalBpiLib.h"
#include <string>
#include <pthread.h>
#include <limits.h>

/**
 * \brief Control- and IO interface for a protocol instance.
 *
 * This class represents the control and IO interface for a protocol instance. An instance of a ProtocolInterface
 * is specified by a name (uid) and an absolute path to an executable (runPath). After the initialization of an instance,
 * it has to be explicitly started (ProtocolInterface::start()). This invokes a thread, which
 * 	1. executes the specified executable
 * 	2. intantiates pipes for communication between the thread and the executed child process and
 * 	3. listens for input from and to the child process while the executed child process is alive.
 */
class ProtocolInterface {
public:
	BpiOutStream bpiout; /**< Standard output printer. */
	BpiOutStream bpierr; /**< Error output printer. */

	ProtocolInterface(std::string uid, std::string runPath); /**< Creates a ProtocolInterface with specified uid and runPath. */
	virtual ~ProtocolInterface();

	void start(); /**< Starts the ProtocolInterface by invoking a thread. */
	void stop(); /**< Stops the running thread. */
	int send(std::string s); /**< Sends a message to STDIN of the executed child process. */
	bool isActive(); /**< Checks whether the ProtocolInterface ist still active or was terminated already. */
	std::string getUid(); /**< Returns the unique identifier of the ProtocolInterface respectivly the corresponding protocol. */
	std::string getRunPath();

private:
	std::string _uid; /**< Unique identifier of the protocol. */
	std::string _runPath; /**< Absolute path to the executeable of the protocol. */
	pid_t _pid; /**< Process identifier of the forked child process. */
	bool _active; /**< Indicates whether the interface ist still active or not. */
	pthread_t _piThread; /**< Thread object. */
	int _fdIntfceIn; /**< Ingoing file fescriptor: Other classes -> ProtocolInterface. */
	int _fdIntfceOut; /**< Outgoing file Descriptor: Other classes -> ProtocolInterface. */
	int _fdChildIn; /**< Ingoing file descriptor: ProtocolInterface -> Child. */
	int _fdChildOut; /**< Outgoing file descriptor: Child -> ProtocolInterface. */
	int _fdChildErr; /**< Outgoing file descriptor for errors: Child -> ProtocolInterface.*/

	static void * _threadEntryFunction(void * This); /**< Entry function for thread execution. */
	void _startThread(); /**< Core of the thread. */
	int _executeProtocol(); /**< Creates pipes, forks the thread and executes the child process. */
	int _write(int fileDescriptor, std::string data); /**< Writes data to a file descriptor. */
	int _read(int fileDescriptor, std::string& data); /**< Reads data from a file descriptor. */
	void _handleChildOutput(std::string out); /**< Handles child ouput.*/
	void _handleChildError(std::string err); /**< Handles child error output.*/
};

#endif /* PROTOCOLINTERFACE_H_ */
