package topology;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;

class FabanDescriptionFileHandler {

	private JsonObject fabanDescriptionFile;
	private JsonObject topologyDescription;
	private JsonArray fabanconfigurationDescription;

	private String selectionMode = "first";

	Topology network;

	public FabanDescriptionFileHandler(String pathToFile) {
		fabanDescriptionFile = new JsonObject();

		try {
			JsonParser parser = new JsonParser();
			JsonElement jsonElement = parser.parse(new FileReader(pathToFile));
			fabanDescriptionFile = jsonElement.getAsJsonObject();
		} catch (FileNotFoundException e) {
			System.err.println("File \"" + pathToFile + "\" not found");
		}
	}

	public FabanDescriptionFileHandler(Topology network) {
		this.network = network;
	}

	public boolean isSucessfullyLoaded() {
		return fabanDescriptionFile != null;
	}

	public Topology getNetwork() {
		return network;
	}

	public void setSelectionMode(String s) {
		selectionMode = s;
	}

	public boolean readAndValidateTopology() {
		boolean success = true;
		System.out.println("> Exploring Faban Description File");
		JsonElement _topoloy = searchObject("topology");
		if (_topoloy == null) {
			System.out.println("    Topology NOT FOUND");
			success = false;
		} else if (!_topoloy.isJsonObject()) {
			System.out.println("    Topology FOUND with WRONG FORMAT (topology)");
			success = false;
		} else {
			topologyDescription = (JsonObject) _topoloy;
			String failure = "";
			if (!isArrayOfStrings(searchObject(topologyDescription, "endstations"))) {
				failure += "endstations, ";
			}
			if (!isArrayOfStrings(searchObject(topologyDescription, "bridges"))) {
				failure += "bridges, ";
			}
//			if (!isArrayOfStrings(searchObject(topologyDescription, "links"))) {
//				failure += "links, ";
//			}
			if (failure.length() > 0) {
				failure = failure.substring(0, failure.length() - 2);
				System.out.println("    Topology FOUND with WRONG FORMAT (" + failure + ")");
				success = false;
			} else {
				System.out.println("    Topology FOUND");
			}
		}
		JsonElement _fabanconfig = searchObject("fabanconfiguration");
		if (_fabanconfig == null) {
			System.out.println("    FabanConfiguration NOT FOUND");
		} else if (!_fabanconfig.isJsonArray()) {
			System.out.println("    FabanConfiguration FOUND with WRONG FORMAT");
			success = false;
		} else {
			fabanconfigurationDescription = (JsonArray) _fabanconfig;
			System.out.println("    FabanConfiguration FOUND");
		}
		return success;
	}

	public boolean readFabanConfiguration() {
		System.out.println("> Start Reading FabanConfiguration from inputfile");
		if (network == null) {
			System.out.println("    First read network from json file. Abort");
			return false;
		}
		JsonElement _fcD = searchObject("fabanconfiguration");
		if (_fcD == null || !_fcD.isJsonArray()) {
			System.out.println("    \"fabanconfiguration\" is not a jsonarray");
			return false;
		}
		fabanconfigurationDescription = (JsonArray) _fcD;
		for (JsonElement wavePair : fabanconfigurationDescription) {
			if (!wavePair.isJsonObject()) {
				System.out.println("    \"fabanconfiguration\" has a wrong format");
				return false;
			}
			JsonObject wp = (JsonObject) wavePair;
			if (wp.has("distributingbridge") && wp.has("wave1") && wp.has("wave2")) {

				JsonElement _db = searchObject(wp, "distributingbridge");
				JsonElement _w1 = searchObject(wp, "wave1");
				JsonElement _w2 = searchObject(wp, "wave2");

				if (!_db.isJsonPrimitive() || !_w1.isJsonObject() || !_w2.isJsonObject()) {
					System.out.println("    \"fabanconfiguration\" has a wrong format (db, waves)");
					return false;
				} else {
					Bridge db = network.getBridge(_db.getAsString());
					if (db != null) {
						Wave w1 = createWaveFromJsonObject((JsonObject) _w1, network, db);
						Wave w2 = createWaveFromJsonObject((JsonObject) _w2, network, db);
						if (w1 != null && w2 != null) {
							FabanMetaInformation fmi = new FabanMetaInformation();
							fmi.setW1(w1);
							fmi.setW2(w2);
							network.addFabanMetaInformation(db, fmi);
						} else {
							System.out.println("    Error reading waves from distributing bridge \"" + db.name + "\"");
							return false;
						}
					}
				}

			} else {
				System.out
						.println("    \"fabanconfiguration\" has a wrong format (db or at least one wave is missing)");
				return false;
			}
		}
		System.out.println("    Done!");
		return true;
	}

	public boolean validateFabanConfiguration() {
		System.out.println("> Start validating faban configuration");
		if (network == null) {
			System.out.println("    No network loaded");
			return false;
		}
		if (network.getWaves().size() == 0) {
			System.out.println("    No Faban configuration loaded");
			return false;
		}
		if (network.getWaves().size() != network.getBridges().size()) {
			System.out.println("    Invalid Faban configuration. Too less waves. Abort");
			return false;
		}
		for (Map.Entry<Bridge, FabanMetaInformation> entry : network.getWaves().entrySet()) {
			if (!isValidFabanMetaInformation(entry.getValue())) {
				System.out.println("    Invalid Faban configuration for distributing bridge " + entry.getKey().name);
				return false;
			} else {
				System.out.println("    Valid Faban configuration for distributing bridge " + entry.getKey().name);
			}
		}
		System.out.println("    Done!");
		return true;
	}

	private boolean isValidFabanMetaInformation(FabanMetaInformation fmi) {
		Wave w1 = fmi.getW1();
		Wave w2 = fmi.getW2();
		ArrayList<Bridge> path1;
		ArrayList<Bridge> path2;
		if (w1.getD().name.equals(w2.getD().name) && !w1.getC().name.equals(w2.getC().name)) {
			Bridge c1 = w1.getC();
			Bridge c2 = w2.getC();
			for (Bridge b : network.getBridges()) {
				path1 = getPath(w1, c1, b);
				path2 = getPath(w2, c2, b);
				if (!disjointPaths(path1, path2, b))
					return false;
			}
		} else {
			return false;
		}
		return true;
	}

	private ArrayList<Bridge> getPath(Wave w, Bridge from, Bridge to) {
		ArrayList<Bridge> path = new ArrayList<>();
		if (from.equals(to)) {
			path.add(from);
			return path;
		} else {
			return getPath(w, path, from, to, from);
		}
	}

	private ArrayList<Bridge> getPath(Wave w, ArrayList<Bridge> path, Bridge from, Bridge to, Bridge add) {
		ArrayList<Bridge> tempPath = new ArrayList<>(path);
		tempPath.add(add);
		ArrayList<Bridge> tempPath2;
		for (DirectedLink l : w) {
			if (l.a.equals(add) && !path.contains(l.b)) {
				if (l.b.equals(to)) {
					tempPath.add(to);
					return tempPath;
				}
				tempPath2 = getPath(w, tempPath, from, to, (Bridge) l.b);
				if (tempPath2 != null && tempPath2.contains(to)) {
					return tempPath2;
				}
			}
		}
		return null;
	}

	private boolean disjointPaths(ArrayList<Bridge> path1, ArrayList<Bridge> path2, Bridge target) {
		if (path1 == null | path2 == null)
			return false;
		for (Bridge b : path1) {
			if (!target.equals(b) && path2.contains(b))
				return false;
		}
		return true;
	}

	private Wave createWaveFromJsonObject(JsonObject wave, Topology network, Bridge d) {
		if (network == null || wave == null || d == null)
			return null;
		Wave w = new Wave(network);
		w.setD(d);
		if (wave.has("checkingbridge") && wave.has("routing")) {
			JsonElement _cb = wave.get("checkingbridge");
			JsonElement _routing = wave.get("routing");
			if (_cb.isJsonPrimitive() && _routing.isJsonObject()) {
				w.setC(network.getBridge(_cb.getAsString()));
				JsonObject routing = (JsonObject) _routing;
				for (Map.Entry<String, JsonElement> entry : routing.entrySet()) {
					Bridge from = network.getBridge(entry.getKey());
					if (from == null) {
						System.out.println("    Invalid routing information in distributing bridge " + d.name);
						return null;
					}
					if (entry.getValue().isJsonArray()) {
						for (JsonElement e : (JsonArray) entry.getValue()) {
							Bridge to;
							if (e.isJsonPrimitive() && (to = network.getBridge(e.getAsString())) != null) {
								w.add(new DirectedLink(from, to));
							} else {
								System.out.println("    Invalid routing information in distributing bridge " + d.name);
								return null;
							}
						}
					} else {
						System.out.println("    Invalid routing information in distributing bridge " + d.name);
						return null;
					}
				}
			} else {
				System.out.println(
						"    Invalid format of \"checkingbridge\" or \"routing\" for distributing bridge " + d.name);
				return null;
			}

		} else {
			System.out.println("    Invalid Wave for distributing bridge " + d.name
					+ ". \"checkingbridge\" or \"routing\" missing.");
		}
		return w;
	}

	public JsonElement searchObject(String key) {
		return searchObject(fabanDescriptionFile, key);
	}

	private JsonElement searchObject(JsonObject root, String key) {
		for (Map.Entry<String, JsonElement> entry : root.entrySet()) {
			if (entry.getKey().equals(key))
				return entry.getValue();
			else if (entry.getValue().isJsonObject()) {
				JsonElement sub = searchObject((JsonObject) entry.getValue(), key);
				if (sub != null)
					return sub;
			}
		}
		return null;
	}

	private boolean isArrayOfStrings(JsonElement e) {
		if (!e.isJsonArray())
			return false;
		for (JsonElement e2 : (JsonArray) e) {
			if (!e2.isJsonPrimitive())
				return false;
		}
		return true;
	}

	public boolean createNetworkFromTopologyDescription() {
		System.out.println("> Creating Network from Topology description");
		network = new Topology();
		JsonArray arrB = null, arrE = null, arrL = null;
		for (Map.Entry<String, JsonElement> entry : topologyDescription.entrySet()) {
			if (entry.getKey().equals("endstations") || entry.getKey().equals("bridge")
					|| entry.getKey().equals("links")) {
				if (!entry.getValue().isJsonArray()) {
					System.out.println("Error creating topology: \"" + entry.getValue() + "\" is not a JsonArray");
					return false;
				}
			}
			switch (entry.getKey()) {
			case "endstations":
				arrE = (JsonArray) entry.getValue();
				break;
			case "bridges":
				arrB = (JsonArray) entry.getValue();
				break;
			case "links":
				arrL = (JsonArray) entry.getValue();
				break;
			default:
				break;
			}
		}
		if (arrE != null && arrB != null && arrL != null) {
			for (JsonElement e : arrE)
				network.addEndStation(e.getAsString());
			System.out.println("    EndStations created");
			for (JsonElement e : arrB)
				network.addBridge(e.getAsString());
			System.out.println("    Bridges created");
			for (JsonElement e : arrL) {
				if (e.isJsonPrimitive()) {
					String l = e.getAsString();
					int divider = l.indexOf('-');
					if (network.addLink(l.substring(0, divider), l.substring(divider + 1)) == false) {
						System.out.println("    Failure adding link " + l.substring(0, divider) + "<->"
								+ l.substring(divider + 1) + ". Aborting");
						network = null;
						return false;
					}
				} else if (e.isJsonObject()) {
					try {
						String a = e.getAsJsonObject().getAsJsonObject("A").get("name").getAsString();
						String b = e.getAsJsonObject().getAsJsonObject("B").get("name").getAsString();
						if (network.addLink(a, b) == false) {
							System.out.println("    Failure adding link " + a + "<->" + b + ". Aborting");
							network = null;
							return false;
						}
					} catch (Exception e2) {
						System.out.println("    Exception occured. Aborting");
						network = null;
						return false;
					}
				}
			}
			System.out.println("    Links created");
		}
		return true;
	}

	public void printBridges() {

		for (Map.Entry<String, JsonElement> entry : fabanDescriptionFile.entrySet()) {
			System.out.println(entry.getKey());
			if (entry.getValue().isJsonObject()) {
				for (Map.Entry<String, JsonElement> entry2 : ((JsonObject) entry.getValue()).entrySet())
					System.out.println(entry2.getKey());
			}
		}
	}

	public void add(String key, String value) {
		fabanDescriptionFile.addProperty(key, value);
	}

	public void save(String path) {
		System.out.println("> Saving json file to \"" + path + "\"");
		try {
			PrintWriter out = new PrintWriter(new File(path));
			Gson g = new GsonBuilder().setPrettyPrinting().create();
			g.toJson(fabanDescriptionFile, out);
			out.close();
			System.out.println("    Done!");
		} catch (FileNotFoundException e) {
			System.out.println("    Saving not sucessfull!");
		}
	}

	public void createJsonWaveObject() {
		System.out.println("> Creation of json fabanconfiguration started");
		if (fabanconfigurationDescription != null) {
			System.out.println("    FabanConfiguration already exists. Overwrite...");
		}
		fabanconfigurationDescription = new JsonArray();
		for (Map.Entry<Bridge, FabanMetaInformation> fmi : network.getWaves().entrySet()) {
			JsonObject jsonWavePair = new JsonObject();
			jsonWavePair.addProperty("distributingbridge", fmi.getValue().getW1().getD().name);

			JsonObject jsonWave1 = new JsonObject();
			jsonWave1.addProperty("checkingbridge", fmi.getValue().getW1().getC().name);
			JsonObject routing1 = new JsonObject();
			JsonObject jsonWave2 = new JsonObject();
			jsonWave2.addProperty("checkingbridge", fmi.getValue().getW2().getC().name);
			JsonObject routing2 = new JsonObject();
			for (Bridge b : network.getBridges()) {
				JsonArray targets1 = new JsonArray();
				for (DirectedLink l : fmi.getValue().getW1()) {
					if (l.a.equals(b)) {
						targets1.add(l.b.name);
					}
				}
				JsonArray targets2 = new JsonArray();
				for (DirectedLink l : fmi.getValue().getW2()) {
					if (l.a.equals(b)) {
						targets2.add(l.b.name);
					}
				}
				routing1.add(b.name, targets1);
				routing2.add(b.name, targets2);
			}
			jsonWave1.add("routing", routing1);
			jsonWavePair.add("wave1", jsonWave1);
			jsonWave2.add("routing", routing2);
			jsonWavePair.add("wave2", jsonWave2);

			fabanconfigurationDescription.add(jsonWavePair);
		}
		if (fabanDescriptionFile.has("fabanconfiguration")) {
			fabanDescriptionFile.remove("fabanconfiguration");
		}
		fabanDescriptionFile.add("fabanconfiguration", fabanconfigurationDescription);
		System.out.println("    Done!");
	}

	public boolean generateWavesSilent() {
		if (!network.isConnected()) {
			return false;
		}
		int i = 0;
		int p = 0;
		int q = 25;
		for (Bridge db : network.getBridges()) {
			if (network.hasWave(db)) {
			} else {
				if (!generateWaves(db)) {
					return false;
				}
				i++;
				int newP = (int) (q * i / ((double) network.getBridges().size())) - p;
				p += newP;
				for (int l = 0; l < newP; l++) {
				}
			}
		}
		return true;
	}

	public boolean generateWaves() {
		System.out.println("> Generation of Waves started");
		if (!network.isConnected()) {
			System.out.println("    Network is not connected. Abort...");
			return false;
		}
		int i = 0;
		System.out.println("       _________________________|100%");
		int p = 0;
		int q = 25;
		System.out.print("    0%|");
		for (Bridge db : network.getBridges()) {
			if (network.hasWave(db)) {
				System.out.println("    Wave for " + db.name + " already exists.\tProgress:  " + ++i + "/"
						+ network.getBridges().size());
			} else {
				if (!generateWaves(db)) {
					System.out.println("    Could not generate waves distributed by bridge \'" + db.name + "\'");
					return false;
				}
				i++;
				int newP = (int) (q * i / ((double) network.getBridges().size())) - p;
				p += newP;
				for (int l = 0; l < newP; l++) {
					System.out.print("X");
				}
			}

		}
		System.out.println("\n    Done!");

		return true;
	}

	private boolean generateWaves(Bridge db) {
		ArrayList<Bridge> checkingBridgeCandidates = network.getNeighbours(db);
		for (int i = 0; i < checkingBridgeCandidates.size(); i++) {
			for (int j = i + 1; j < checkingBridgeCandidates.size(); j++) {
				Bridge c1 = checkingBridgeCandidates.get(i);
				Bridge c2 = checkingBridgeCandidates.get(j);
				if (generateWaves(db, c1, c2))
					return true;
			}
		}
		return false;
	}

	private boolean generateWaves(Bridge db, Bridge c1, Bridge c2) {
		HashMap<Bridge, Integer> distanceMap = new HashMap<Bridge, Integer>();
		HashMap<Bridge, Integer> distanceMap2 = new HashMap<Bridge, Integer>();

		// FABAN1: Fig.5: (05) - (06)
		Wave w1 = new Wave(network);
		Wave w2 = new Wave(network);
		w1.add(new DirectedLink(db, c1));
		w1.add(new DirectedLink(c1, db));
		distanceMap.put(c1, 1);
		distanceMap.put(db, 2);
		ArrayList<DirectedLink> hatW = new ArrayList<DirectedLink>();
		ArrayList<DirectedLink> hatC = new ArrayList<DirectedLink>();

		Component x, y;
		DirectedLink dl;

		// FABAN1: Fig.5: (08) - (14)
		for (UndirectedLink l : network.getBridgeLinks()) {
			x = l.a;
			y = l.b;
			dl = new DirectedLink(x, y);
			if (!x.equals(db) && !y.equals(db) && !x.equals(c2))
				hatW.add(dl);
			if (!x.equals(db) && !y.equals(db) && !x.equals(c1))
				hatC.add(dl);

			x = l.b;
			y = l.a;
			dl = new DirectedLink(x, y);
			if (!x.equals(db) && !y.equals(db) && !x.equals(c2))
				hatW.add(dl);
			if (!x.equals(db) && !y.equals(db) && !x.equals(c1))
				hatC.add(dl);
		}

		// FABAN1: Fig.5: (15) - (22)
		while (w1.size() < network.getBridges().size() && !w1.refreshNE(hatW).isEmpty()) {
			DirectedLink e = w1.selectRandomEdge(selectionMode);

			if (validLinkSelectionStep(w1, c1, c2, hatC, e)) {
				w1.add(e);
				distanceMap.put((Bridge) e.b, distanceMap.get(e.a) + 1);
				hatC.remove(e);
			}
			hatW.remove(e);
		}

		// FABAN1: Fig.5: (23) - (26)
		if (w1.size() == network.getBridges().size()) {
			w1.setD(db);
			w1.setC(c1);

		} else
			return false;

		int max = 0;
		for (Integer i : distanceMap.values()) {
			if (max < i)
				max = i;
		}
		w1.setMaxDistance(max);

		// FABAN1: Fig.8: (05)
		w2.add(new DirectedLink(db, c2));
		w2.add(new DirectedLink(c2, db));
		distanceMap2.put(c2, 1);
		distanceMap2.put(db, 2);

		// FABAN1: Fig.8: (12) - (18)
		while (w2.size() < network.getBridges().size() && !w2.refreshNE(hatC).isEmpty()) {
			DirectedLink e = w2.selectRandomEdge(selectionMode);
			if (validLinkSelectionStep(w2, c2, c1, w1, e)) {
				w2.add(e);
				distanceMap2.put((Bridge) e.b, distanceMap2.get(e.a) + 1);
			}
			hatC.remove(e);
		}

		// FABAN1: Fig.8: (19) - (21)
		if (w2.size() == network.getBridges().size()) {
			w2.setD(db);
			w2.setC(c2);
		} else
			return false;

		int max2 = 0;
		for (Integer i : distanceMap2.values()) {
			if (max2 < i)
				max2 = i;
		}
		w2.setMaxDistance(max2);

		FabanMetaInformation fmi = new FabanMetaInformation();
		fmi.setW1(w1);
		fmi.setW2(w2);
		network.addFabanMetaInformation(db, fmi);

		return true;
	}

	private boolean validLinkSelectionStep(ArrayList<DirectedLink> W, Bridge c1, Bridge c2,
			ArrayList<DirectedLink> paramHatC, DirectedLink e) {
		ArrayList<DirectedLink> hatC = new ArrayList<>(paramHatC);
		hatC.remove(e);

		ArrayList<Bridge> forbiddenBridges = new ArrayList<>();
		Bridge temp = (Bridge) e.a;
		forbiddenBridges.add(temp);
		while (!temp.equals(c1)) {
			for (DirectedLink l : W) {
				if (l.b.equals(temp)) {
					temp = (Bridge) l.a;
					forbiddenBridges.add(temp);
				}
			}
		}
		if (existsPath(c2, (Bridge) e.b, hatC, forbiddenBridges))
			return true;
		return false;
	}

	private boolean existsPath(Bridge from, Bridge to, ArrayList<DirectedLink> linkset,
			ArrayList<Bridge> forbiddenBridges) {
		if (from.equals(to))
			return true;
		ArrayList<Bridge> beginning = new ArrayList<Bridge>();
		beginning.add(from);
		boolean result = existsPath(beginning, from, to, linkset, forbiddenBridges);
		return result;
	}

	private boolean existsPath(ArrayList<Bridge> actPath, Bridge newest, Bridge to, ArrayList<DirectedLink> linkset,
			ArrayList<Bridge> forbiddenBridges) {
		ArrayList<Bridge> tempPath = new ArrayList<>(actPath);
		for (DirectedLink l : linkset) {
			if (l.a.equals(newest) && !forbiddenBridges.contains(l.b) && !tempPath.contains(l.b)) {
				tempPath.add((Bridge) l.b);
				if (l.b.equals(to))
					return true;
				else if (existsPath(tempPath, (Bridge) l.b, to, linkset, forbiddenBridges))
					return true;
			}
		}
		return false;
	}
}
